// =============================================================
// ==
// ==   Dataserver routine: herastatebase.c
// ==
// =============================================================
#include "prolog.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <ctype.h>
#include <signal.h>
#include <dirent.h>
#include "fecerr.h"
#include "rpctypes.h"
#include "database.h"
#include "almsrv.h"
#include "errors.h"
#include "archive.h"
#include "statebase.h"
#include "toolkit.h"
#include "brdcast.h"

extern FEC *fec;

#define HEINJENERGY 12.0
#define HPINJENERGY 40.0
#define ZEUSTOP 3945
#define H1TOP 55.17
float H1Solen = -1;
float ZeusSolen = -1;

NAME16 ZeusCal;
char *ZeusCalEnum[5] = {"Halted","Opening","Closing","Open","Closed" };
#define nZeusCalPos 5
int iZeusCalPos = 0;

// important initialization:
short iHPenergy;
double HPCur = -1,HPTau = -1,HECur = -1,HETau = -1;
float fminus = -1;
float *HPenergy = &fminus,*HEenergy = &fminus;

// state variables:
extern int ramping, running, beam, lumi;
extern int Pramping, Eramping, Pbeam, Ebeam;
extern int Prunning, Erunning;
extern int DclEramping, DclPramping, DclLumi;
extern int DclEInj, DclPInj; 
extern int pneedsToDeclare;
extern int eneedsToDeclare;

#define CHECKSTATES
NAME16 allowedNullStates[] =
{
# ifdef CHECKSTATES
  { "Undefined" },
  { "Access Ready" },
  { "Access" },
  { "Technical Proble" },
  { "Not Operating" },
  { "Idle" },
  { "Off" },
  { "PS Not Ready" },
  { "Massage" },
  { "Massage end" },
  { "Massage Abort" },
# endif
};
int nAllowedNullStates = sizeof(allowedNullStates)/sizeof(NAME16);
 
CHECKNAME checkedParameters[] =
{
# ifdef CHECKSTATES
  { "HPMAGEN", CF_FLOAT, NULL, -1, 1 },
  { "HPDCCur", CF_FLOAT, NULL, -1, 1 },
  { "HEMAGEN", CF_FLOAT, NULL, -1, 0 },
  { "HEDCCur", CF_FLOAT, NULL, -1, 0 },
# endif
};
int nCheckedParameters = sizeof(checkedParameters)/sizeof(CHECKNAME);

CHECKNAME checkedStates[] =
{
# ifdef CHECKSTATES
  { "HPSTATE", CF_NAME16, NULL, -1, 1 },
  { "HESTATE", CF_NAME16, NULL, -1, 0 },
# endif
};
int nCheckedStates = sizeof(checkedStates)/sizeof(CHECKNAME);
 
char Eproc[32] = "none", Pproc[32] = "none";
char Etype[32] = "Elektronen", Ptype[32] = "Protonen";
 
DECLPROC eDeclaredProcedures[] =
{
  {"","","heundefined"},
  {"","","heinjection"},
};
int nEDeclaredProcedures = sizeof(eDeclaredProcedures)/sizeof(DECLPROC);
DECLPROC pDeclaredProcedures[] =
{
  {"","","hpundefined"},
  {"","","hpinjection"},
};
int nPDeclaredProcedures = sizeof(pDeclaredProcedures)/sizeof(DECLPROC);
 
char *gTargetStateServer = "/HERA/HERASTATE";
 
extern int BroadcastServer;
extern SRVINFO srvinfo;
extern int globalsWatchdog, globalsServer;
int getStartupInfo(void);
 
void PreSystemInit(void)
{
  if (getStartupInfo()) exit(2);
  PortOffset = srvinfo.port;
  BroadcastServer = TRUE;
  GlobalHeartbeat = 3;
# ifdef GLOBALS_SERVER  
  globalsServer = TRUE;
# else
  globalsWatchdog = TRUE;
# endif
  MaxNumClients = 500;
  RegisterFecNameEx(srvinfo.name,"[SER]Globals Server","Linux","Bldg 30 Rm 103","none","P.Duval",srvinfo.port,"HERA");
  FeclogDepth = 2000;
  return;
}
char *trimString(char *str)
{
  static char trimBuffer[256];
  char *pstr;
  if (str == NULL)
  {
    trimBuffer[0] = 0;
    goto err;
  }
  for (pstr=str; strchr(" \t\n",*pstr) != NULL; pstr++);
  strncpy(trimBuffer,str,256);
  strtok(trimBuffer," \t\n");
err:
  return trimBuffer;
}

void updateStateInfo(int i, int j,int cc)
{
  static int lastPEnergy;
  static float lastEEnergy;
  // monitor crucial parameters ...
  static int PEpanic = 0, EEpanic = 0,PIpanic = 0; 
  static int STpanic = 0;
  static int H1panic = 0;
  static int ZEpanic = 0;
  static int iPEpanic = 0;
  float *fptr;
  int k;
  
  if (!strncmp(fec[i].fecdata[j].param,"HEbpm",5) || !strncmp(fec[i].fecdata[j].param,"HPbpm",5))
  {
    fptr = (float *)(fec[i].data+fec[i].fecdata[j].offset);
    for (k=0; k<fec[i].fecdata[j].size; k++)
    {
      if (*fptr > 1e11) *fptr = 0;
      fptr++;
    }
    goto fixstate;
  }
  
  if (!strcmp(fec[i].fecdata[j].param,"HPENERGY"))
  {
    iHPenergy = *(short *)(fec[i].data+fec[i].fecdata[j].offset);
    if (cc)            // there's something wrong
    {
      if (++iPEpanic > 5)
      {
        iHPenergy = -1;
      }
    }
    else               // everything's okay
    {
      iPEpanic = 0;
    }
    goto fixstate;
  }

  if (!strcmp(fec[i].fecdata[j].param,"HPMAGEN"))
  {
    HPenergy = (float *)(fec[i].data+fec[i].fecdata[j].offset);
    if (cc)            // there's something wrong
    {
      if (++PEpanic > 5)
      {
        *HPenergy = -1;
        SetAlarm("FILER",0,hpenergy_undefined,NULL);
        if (PEpanic == 6) feclog("Missing P energy");
      }
    }
    else               // everything's okay
    {
      if (abs(*HPenergy - lastPEnergy) > 1)
      {
        Pramping = (int)(*HPenergy - lastPEnergy);
        lastPEnergy = *HPenergy;
      }
      else Pramping = 0;
      PEpanic = 0;
      ClearAlarm("FILER",0);            
    }
//    iHPenergy = (int)*HPenergy;
    Prunning =  *HPenergy > 2 ? 1 : 0;
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HEMAGEN"))
  {
    HEenergy = (float *)(fec[i].data+fec[i].fecdata[j].offset);
    if (cc)          // there's something wrong
    {
      if (++EEpanic > 5)
      {
        *HEenergy = -1;
        SetAlarm("FILER",1,heenergy_undefined,NULL);
        if (EEpanic == 6) feclog("Missing E energy");
      }
    }
    else             // everything's okay
    {
      if (abs(*HEenergy - lastEEnergy) > 1.0)            
      {
        Eramping = (int)(*HEenergy - lastEEnergy);
        lastEEnergy = *HEenergy;
      }
      else Eramping = 0;
      EEpanic = 0;
      ClearAlarm("FILER",1);
    }
    Erunning =  (*HEenergy > 1) ? 1 : 0;
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HPDCCur"))
  {
    HPCur = (double)(*(float *)(fec[i].data+fec[i].fecdata[j].offset));
    if (cc && ++PIpanic < 6) goto fixstate;
    if (cc) HPCur = -1.0;
    Pbeam = HPCur > 0.1 ? 1 : 0;
    if (HPCur != -1 && HPCur < 0) HPCur = 0;
    PIpanic = 0; 
    if (HPCur == -1) SetAlarm("FILER",2,hpcurrent_undefined,NULL);
    else ClearAlarm("FILER",2);
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HPDCLoss"))
  {
    HPTau = (double)(*(float *)(fec[i].data+fec[i].fecdata[j].offset));
    if (cc && ++PIpanic < 6) goto fixstate;
    HPTau = cc ? -1 : fabs(HPTau);
    if (HPTau < 0.0001) HPTau = 0.0001;
    HPTau = 1/HPTau;
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HEDCCur"))
  {
    HECur = (double)(*(float *)(fec[i].data+fec[i].fecdata[j].offset));
    if (cc && ++PIpanic < 6) goto fixstate;
    if (cc) HECur = -1;
    Ebeam = HECur > 0.1 ? 1 : 0; 
    if (HECur != -1 && HECur < 0) HECur = 0;
    if (HECur == -1)SetAlarm("FILER",3,hecurrent_undefined,NULL);
    else ClearAlarm("FILER",3);
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HEDCLoss"))
  {
    HETau = (double)(*(float *)(fec[i].data+fec[i].fecdata[j].offset));
    if (cc && ++PIpanic < 6) goto fixstate;
    HETau = cc ? -1 : fabs(HETau);
    if (HETau < 0.0001) HETau = 0.0001;
    HETau = 1/HETau;
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HESTATE"))
  {
    if (cc && ++STpanic < 6) goto fixstate;
    if (cc) { DclEInj = DclEramping = DclLumi = 0; goto fixstate; }
    STpanic = 0;
    DclEramping = !strcmp("Ramp",fec[i].data+fec[i].fecdata[j].offset);
    DclEInj = !strcmp("Injection Ready",fec[i].data+fec[i].fecdata[j].offset);
    DclLumi = !strcmp("Lumi Run",fec[i].data+fec[i].fecdata[j].offset);
    // check Injection ready following Massage End ...
    if (!strcmp("Massage end",fec[i].data+fec[i].fecdata[j].offset))
    {
      if (fabs(*HEenergy-HEINJENERGY) < 0.5)
      {
        e_stateindex = 1;  // hardwired to Injection Ready
        eneedsToDeclare = TRUE;
        return;
      }
    }
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"HPSTATE"))
  {
    if (cc && ++STpanic < 6) goto fixstate;
    if (cc) { DclPInj = DclPramping = DclLumi = 0; goto fixstate; }
    STpanic = 0;
    DclPramping = !strcmp("Ramp",fec[i].data+fec[i].fecdata[j].offset);
    DclPInj = !strcmp("Injection Ready",fec[i].data+fec[i].fecdata[j].offset);
    DclLumi = !strcmp("Lumi Run",fec[i].data+fec[i].fecdata[j].offset);
    // check Injection ready following Massage End ...
    if (!strcmp("Massage end",fec[i].data+fec[i].fecdata[j].offset))
    {
      if (fabs(*HPenergy-HPINJENERGY) < 0.5)
      {
        p_stateindex = 1;  // hardwired to Injection Ready
        pneedsToDeclare = TRUE;
        return;
      }
    }
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"H1SolenCur"))
  {
    H1Solen = (*(float *)(fec[i].data+fec[i].fecdata[j].offset));
    if (cc && ++H1panic < 6) goto fixstate;
    if (cc) H1Solen = -1; else H1panic = 0;
    if (H1Solen != -1) H1Solen /= H1TOP;
    if (H1Solen == -1) SetAlarm("FILER",3,h1solen_undefined,NULL);
    else ClearAlarm("FILER",3);
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"ZeusSolen"))
  {
    ZeusSolen = (*(float *)(fec[i].data+fec[i].fecdata[j].offset));
    if (cc && ++ZEpanic < 6) goto fixstate;
    if (cc) ZeusSolen = -1; else ZEpanic = 0;
    if (ZeusSolen != -1)
    {
      ZeusSolen = (float)((int)(ZeusSolen + .5) * 100/ ZEUSTOP);
      (*(float *)(fec[i].data+fec[i].fecdata[j].offset)) = ZeusSolen;
    }
    if (ZeusSolen == -1) SetAlarm("FILER",3,ZeusSolen_undefined,NULL);
    else ClearAlarm("FILER",3);
    goto fixstate;
  }
  if (!strcmp(fec[i].fecdata[j].param,"ZeusCalPos"))
  {
    iZeusCalPos = (*(int *)(fec[i].data+fec[i].fecdata[j].offset));
    iZeusCalPos = iZeusCalPos % nZeusCalPos;
    if (cc && ++ZEpanic < 6) goto fixstate;
    if (cc) iZeusCalPos = -1; else ZEpanic = 0;
    if (iZeusCalPos != -1)
    {
      strncpy(ZeusCal.name,ZeusCalEnum[iZeusCalPos],16);
      ClearAlarm("FILER",3);
    }
    else
    {
      strncpy(ZeusCal.name,"-1",16);
      SetAlarm("FILER",3,ZeusCal_undefined,NULL);
    }
    goto fixstate;
  }

fixstate:
  lumi = DclLumi;
  running = Erunning | Prunning;
  beam = Ebeam | Pbeam;
  ramping = DclEramping | DclPramping; 
  // HERA sequencer otherwise will take care of declaring states ...
  return;
  for (i=0; i<nEDeclaredProcedures; i++)
  {
    if (strlen(eDeclaredProcedures[i].typestr) > 0 &&
        stricmp(Etype,eDeclaredProcedures[i].typestr) != 0) continue;
    if (!stricmp(Eproc,eDeclaredProcedures[i].procstr)) break;
  }
  e_stateindex = (i < nEDeclaredProcedures) ? i : 0;
  for (i=0; i<nPDeclaredProcedures; i++)
  {
    if (strlen(pDeclaredProcedures[i].typestr) > 0 &&
        stricmp(Ptype,pDeclaredProcedures[i].typestr) != 0) continue;
    if (!stricmp(Pproc,pDeclaredProcedures[i].procstr)) break;
  }
  p_stateindex = (i < nPDeclaredProcedures) ? i : 0;
}

int overloadEqpmod(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc, i, j, n, len;
  char prop[32], *c;
  short stateArray[7];
  short dsize,EqNo = 0,fmt;                                                     

  if (devName[0] == '#') EqNo = atoi(&devName[1]);
  if (EqNo < 0) return illegal_equipment_number;
  if (EqNo > 0) EqNo = EqNo - 1;                  // count from 0
  fmt = dout->dFormat;
  dsize = dout->dArrayLength;              

  if (!strcmp(devProperty,"H1Solen"))
  {
    return putValuesFromFloat(dout,&H1Solen,1);
  }  
  if (!strcmp(devProperty,"ZeusCal"))
  {
    if (dout->dArrayLength < 1) return dimension_error;
    switch (dout->dFormat)
    {
      case CF_TEXT:
        len = dout->dArrayLength;
        break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
        len = fmtsizeof(dout->dFormat);
        break;
      default:
        return illegal_format;
    }
    if (len > 16) len = 16;
    strncpy(dout->data.cptr,ZeusCal.name,len);
    return 0;
  }  
  if (!strcmp(devProperty,"HEbpmXH1"))
  {
    strncpy(devProperty,"HEbpmX",32);
    sprintf(devName,"#%d",103);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmYH1"))
  {
    strncpy(devProperty,"HEbpmY",32);
    sprintf(devName,"#%d",103);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmXB"))
  {
    strncpy(devProperty,"HEbpmX",32);
    sprintf(devName,"#%d",45);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmYB"))
  {
    strncpy(devProperty,"HEbpmY",32);
    sprintf(devName,"#%d",45);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmXHE"))
  {
    strncpy(devProperty,"HEbpmX",32);
    sprintf(devName,"#%d",171);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmYHE"))
  {
    strncpy(devProperty,"HEbpmY",32);
    sprintf(devName,"#%d",171);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmXZU"))
  {
    strncpy(devProperty,"HEbpmX",32);
    sprintf(devName,"#%d",242);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HEbpmYZU"))
  {
    strncpy(devProperty,"HEbpmY",32);
    sprintf(devName,"#%d",242);
    return illegal_property; // force call further down stream
  } 
  if (!strcmp(devProperty,"HPENERGY"))
  {
    if (dsize != 1) return illegal_read_write;
    return putValuesFromShort(dout,&iHPenergy,1);
    return 0;
  }  
  if (!stricmp(devProperty,"HECur") || !strcmp(devProperty,"HESUCURR") ||
      !strcmp(devProperty,"HECurMKI"))
  {
    if (dsize != 1) return illegal_read_write;
    if (fmt == CF_FLOAT) dout->data.fptr[0] = HECur;
    else if (fmt == CF_DOUBLE) dout->data.dptr[0] = HECur;
    else return illegal_format;
    return 0;
  }
  if (!stricmp(devProperty,"HETau") || !strcmp(devProperty,"HESULIFE") ||
      !strcmp(devProperty,"HEtauMKI") || !strcmp(devProperty,"HEBuntau"))
  {
    if (dsize != 1) return illegal_read_write;
    if (fmt == CF_FLOAT) dout->data.fptr[0] = HETau;
    else if (fmt == CF_DOUBLE) dout->data.dptr[0] = HETau;
    else return illegal_format;
    return 0;
  } 
  if (!stricmp(devProperty,"HPCur") || !strcmp(devProperty,"HPSUCURR") ||
      !strcmp(devProperty,"HPCurMKI"))
  {
    if (dsize != 1) return illegal_read_write;
    if (fmt == CF_FLOAT) dout->data.fptr[0] = HPCur;
    else if (fmt == CF_DOUBLE) dout->data.dptr[0] = HPCur;
    else return illegal_format;
    return 0;
  }
  if (!stricmp(devProperty,"HPTau") || !strcmp(devProperty,"HPSULIFE") ||
      !strcmp(devProperty,"HPtauMKI") || !strcmp(devProperty,"HPBuntau"))
  {
    if (dsize != 1) return illegal_read_write;
    if (fmt == CF_FLOAT) dout->data.fptr[0] = HPTau;
    else if (fmt == CF_DOUBLE) dout->data.dptr[0] = HPTau;
    else return illegal_format;
    return 0;
  }  
  if (!strcmp(devProperty,"HEinjSt") || !strcmp(devProperty,"HPinjSt"))
  {
    return 0;
  }
  return illegal_property;
}
extern void cbackBrdsrv(int id,int cc);
int GlobalsRegistrationEx(void)
{
  DTYPE din;   
  // for historical reasons:
  din.dFormat = CF_SHORT;
  din.dArrayLength = 1;
  din.data.vptr = &iHPenergy;
  sendNetGlobal("HPENERGY",&din,cbackBrdsrv,1,1,0);
  din.dFormat = CF_DOUBLE;
  din.dArrayLength = 1;
  din.data.vptr = &HPCur;
  sendNetGlobal("HPCur",&din,cbackBrdsrv,1,1,0);
  din.data.vptr = &HECur;
  sendNetGlobal("HECur",&din,cbackBrdsrv,1,1,0);
  din.data.vptr = &HPTau;
  sendNetGlobal("HPtau",&din,cbackBrdsrv,1,1,0);
  din.data.vptr = &HETau;
  sendNetGlobal("HEtau",&din,cbackBrdsrv,1,1,0);  
  din.dFormat = CF_FLOAT;
  din.data.vptr = &H1Solen;
  sendNetGlobal("H1Solen",&din,cbackBrdsrv,1,1,0);  
  din.dFormat = CF_NAME16;
  din.data.vptr = &ZeusCal;
  sendNetGlobal("ZeusCal",&din,cbackBrdsrv,1,1,0);  
  return 0;
}

