//#include <cstdio>
//#include <cstdlib>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "mstlib.h"
#include "ecc_helpers.h"
 
#if defined(ATTOCOM_EXPORTS) 
# define LIB_EXPORT __declspec(dllexport)
# else
# define LIB_EXPORT __declspec(dllimport)
# endif

#define CONFIG_FILE "attousb.csv"
#define BIN_PATH "."
 
#define CLBR_PATTERN 32641
#define MOVETO_DURATION 600
 
NAME32 stsLst[] = {
  "moving",
  "paused",
  "",
  "",
  "calibrating (step1)",
  "calibrating (step2)",
  "calibrating (step3)",
  "",
  "",
  "",
  "",
  "is calibrating",
  "uncalibrated",
  "",
  "",
  "hardware error",
};
 
int sizStsLst = sizeof(stsLst)/sizeof(NAME32);
 
#define MTR_MIN_POSITION -128000000
 
#define MTR_TYPE_UNKNOWN  0x00
#define MTR_TYPE_LINEAR   0x01
#define MTR_TYPE_ROTATION 0x02
#define MTR_TYPE_GONIOMETER 0x03

int attoDebugLevel = 0;
int getNmPerStep(int mot,int *nmps);

int GetMotorResponse(int mot,const char *inpt,char *outp,int outlen)
{
  int cc=0, p=0, slen=strlen(inpt);

  if (attoDebugLevel && cc) dbglog("motor %d: <%d>",mot,cc);
  return cc;
}
int getMotorType(char *typeStr)
{
  if (typeStr == NULL) return MTR_TYPE_UNKNOWN;
  if (!stricmp(typeStr,"LINEAR")) return MTR_TYPE_LINEAR;
  if (!stricmp(typeStr,"ROTATION")) return MTR_TYPE_ROTATION;
  if (!stricmp(typeStr,"GONIOMETER")) return MTR_TYPE_GONIOMETER;
 return MTR_TYPE_UNKNOWN;
}
 
#define STS_BUFFER_SIZE 20
#define PSAX_BUFFER_SIZE 20

typedef struct
{
  double rbBuffer[PSAX_BUFFER_SIZE];
  double tgtBuffer[PSAX_BUFFER_SIZE];
  int stsBuffer[PSAX_BUFFER_SIZE];
  int grpIdXref[PSAX_BUFFER_SIZE];
  int status[PSAX_BUFFER_SIZE];
  int lnkStsPosPsAx; 
  int lnkIdPosPsAx; 
  int lnkStsStsPsAx; 
  int lnkIdStsPsAx;
  int lnkIdTgtPsAx;
  int lnkStsTgtPsAx;
  int numPsAxGroups;
  int idLastScanned;
} PseudoAxisGroups;

PseudoAxisGroups psAxGrps;

typedef struct
{
  int id;
  float gearRatio;
  MotorGroup grp;
} MotGrpTbl;
#define GRPTBL_SIZE 10
MotGrpTbl grps[GRPTBL_SIZE];
int ngrps = 0;

// status bit definitions (property "StatusInfo" -> stsBits)
#define STS_IDLE 0x01
#define STS_ENABLED 0x02
#define STS_MOVING 0x04
#define STS_MOVPOS 0x08
#define STS_MOVNEG 0x10
#define STS_ENDPOS 0x20
#define STS_ENDNEG 0x40
#define STS_OVERRIDE 0x80

#define FREQ_MIN 1
#define FREQ_MAX 10000
#define VOLT_MIN 0
#define VOLT_MAX 150

typedef struct
{
  char name[32];
  char target[128];
  char units[16];
  char id[32];
  int type;
  int axis;
  int grpId;   // -1 if not currently a group member
  int idle;    // bit
  int enabled; // bit
  int moving;  // bit
  int stepping;
  int nsteps;
  int movpos;  // bit
  int movneg;  // bit
  int endpos;  // bit
  int endneg;  // bit
  int axisState; // word ? byte ?
  int coupledState;// word ? byte ?
  int clbrState;// current calibration state in calibration procedure
  int clbrSts;  // is calibrated or not
  int actCntrlLoopIndex;// word ? byte ?
  int frequency;
  float voltage;
  float voltageRb;
  float nmPerStep;
  float stsBuffer[STS_BUFFER_SIZE];
  double pos;
  double posRb; // readback position
  double tgt;
  double tgtRb; // readback target position
  double posRef; // reference position
  double moveStart;
  float overrideSts;
  float lag;  // soll - ist
  float velRb;
  float pcntMaxCur;
  float maxPos; // maximum position (micro steps)
  float minPos; // minimum position (micro steps)
  float offPos; // offset position (micro steps)
  float scalePos; // scale factor (usually -1.0 or 1.0)
  float velMean; // normal velocity (msteps / sec)
  float velLwr;
  float velUpr;
  float accUpr;
  float rbTolerance;
  float masterGearRatio; // 1.0 default
  int movCmd;
  int movEtr; // 'estimated time remaining' for move action
  int movPaused;
  int status;
  int cmdSts;
  int linkId;
  THRHANDLE hIoThread;
  float baudRate;
  int servoOn;
  int driftComp;
  char rbBuf[64];
  time_t lastCheckedClbr;
} CommMotor;
 
CommMotor *motors = NULL;
 
int gNumMotors = 1;
void cbTgtPos(int id, int cc);
void startUpdateMotor(int mot);
LIB_EXPORT int moveto(int mot,int cmd);
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd);
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp);
LIB_EXPORT int grpmoveto(int grpId,int cmd);
 
typedef struct
{
  char target[128];
  char name[64];
  char type[32];
  char units[16];
  int motorNr;
  int axis;
  int frequency;
  float voltage;
  float posMax;
  float posMin;
  float posOff;
  float posScale;
  float rbTolerance;
  float gearRatio;
  float nmStep;
} MotorInfoStruct;
 
CSVCOLUMN csvCommConfigColumns[] =
{
  { "TARGET","",0,CF_TEXT,128,(BYTE *)OFFSETIN(MotorInfoStruct,target),NULL,0 },
  { "NAME","",0,CF_TEXT,64,(BYTE *)OFFSETIN(MotorInfoStruct,name),NULL,0 },
  { "TYPE","LINEAR",0,CF_TEXT,32,(BYTE *)OFFSETIN(MotorInfoStruct,type),NULL,0 },
  { "MOTOR","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,motorNr),NULL,0 },
  { "AXIS","",1,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,axis),NULL,0 },
  { "UNITS","nm",0,CF_TEXT,16,(BYTE *)OFFSETIN(MotorInfoStruct,units),NULL,0 },
  { "NM_STEP","12800",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,nmStep),NULL,0 },
  { "POSITION_MAXIMUM","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMax),NULL,0 },
  { "POSITION_MINIMUM","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMin),NULL,0 },
  { "POSITION_OFFSET","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posOff),NULL,0 },
  { "POSITION_SCALE","1",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posScale),NULL,0 },
  { "POSITION_TOLERANCE","0.01",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,rbTolerance),NULL,0 },
  { "FREQUENCY","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,frequency),NULL,0 },
  { "VOLTAGE","30",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,voltage),NULL,0 },
  { "GEAR_RATIO","1.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,gearRatio),NULL,0 },
};
CSVDB csvCommConfigDb =
{
  csvCommConfigColumns,
  sizeof(csvCommConfigColumns)/sizeof(csvCommConfigColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(MotorInfoStruct),
  0
};

NAME64 *mstNames = NULL;
 
int getPsAxGroupIndex(int grpId)
{
  int i;
  for (i=0; i<psAxGrps.numPsAxGroups; i++)
  {
    if (psAxGrps.grpIdXref[i] == grpId) return i;
  }
  return -1;
}
static MotorInfoStruct *motorInfo = NULL;

#ifdef NO_ENCODER
typedef struct
{
  int mot;
  float pos;
} MotorPositionInfo;
MotorPositionInfo *motPosInf = NULL;

#define MOTOR_POSITIONS_FILE "motor-positions.csv"
CSVCOLUMN csvMotPosColumns[] =
{
  { "MOTOR","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorPositionInfo,mot),NULL,0 },
  { "VALUE","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorPositionInfo,pos),NULL,0 },
};
CSVDB csvMotPosConfigDb =
{
  csvMotPosColumns,
  sizeof(csvMotPosColumns)/sizeof(csvMotPosColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(MotorPositionInfo),
  0
};
int eagerSaveRestore = TRUE;
int saveMotorPostion(int mot)
{ /* input argument doesn't play a role here */
  int i, cc = 0;
  for (i=0; i<csvMotPosConfigDb.siz; i++)
  {
    if (motPosInf[i].mot < 0 || motPosInf[i].mot >= gNumMotors) continue;
    motPosInf[i].pos = (float)motors[motPosInf[i].mot].pos;
  }
  if ((cc=csvWriteFile(NULL,MOTOR_POSITIONS_FILE,&csvMotPosConfigDb)) != 0) ccerr(cc);

err:
  if (cc != 0) feclog("could not save motor %d position : %.32s",mot,cc2str(cc));
  return cc;
}
int restoreMotorPosition(int mot)
{
  int i, cc = 0, createFile=FALSE;
retryReadFile:
  if ((cc=csvReadFile(NULL,MOTOR_POSITIONS_FILE,&csvMotPosConfigDb,(void **)&motPosInf)) != 0) 
  {
    if (cc == no_such_file && !createFile)
    { // make the initial file
      FILE *fp = fopen(MOTOR_POSITIONS_FILE,"w");
      if (fp == NULL) ccerr(file_error);
      fprintf(fp,"MOTOR, VALUE\n");
      for (i=0; i<gNumMotors; i++) fprintf(fp,"%d, %g\n",i,motors[i].pos);
      if (fp != NULL) fclose(fp);
      createFile = TRUE;
      goto retryReadFile;
    }
    ccerr(cc);
  }
  if (createFile) ccerr(cc);
  for (i=0; i<csvMotPosConfigDb.siz; i++)
  {
    if (motPosInf[i].mot < 0 || motPosInf[i].mot >= gNumMotors) continue;
    motors[motPosInf[i].mot].pos = (double)motPosInf[i].pos;
  }
err:
  if (cc) feclog("could not restore motor %d position : %.32s",mot,cc2str(cc));
  return cc;
}
#endif
#define MTR_UPDATE_INTERVAL 100
int mstCommTerminating = FALSE;
#ifdef MULTITHREADED
int calibrStep = 0;
THREAD_TYPE motorThreadTask(THREAD_PARAM lpvThreadParam)
{
  int mot = (int)lpvThreadParam;
  int m, cc, pos, moving, tgt, enable, refpos, isMoving;
  Bln32 sts, valid;
  double dpos;
  static int calibrDir = 1; // 1 = forward, 0 = backward

  m = motors[mot].axis;
  // can't read the target setting, so take the position
  // readback at server start:
  //motors[mot].tgt = motors[mot].pos;

  feclog("atto thread monitor motor %d",mot);
  while (!mstCommTerminating)
  {
    Sleep(MTR_UPDATE_INTERVAL);
#   ifdef NO_ENCODER
    if ((motors[mot].status & MTR_MOVE) == 0) 
    { // not moving 
      continue;
    }
#   endif
    if (ECC_getPosition(m,&pos) != 0) continue;
    dpos = (double)pos;
    isMoving = (fabs(dpos - motors[mot].pos) > 100) ? TRUE : FALSE;
    motors[mot].pos = dpos;

    if (motors[mot].status & MTR_UNCLBR)
    {
      if ((cc=ECC_getStatusReference(m,&valid)) != 0)
      {
        motors[mot].status &= ~MTR_UNCLBR; // stop the process
        continue;
      }
      if (!valid)
      {
        if (!isMoving)
        { // not valid and not moving
          if ((calibrStep%2) == 0)
          { // starting off
            enable = 1;
            cc = calibrDir ? ECC_controlContinousFwd(m,&enable,1) : ECC_controlContinousBkwd(m,&enable,1);
            calibrStep++;
          }
          else if (calibrStep == 1)
          { // finished search in one direction and still not valid
            calibrDir = (calibrDir+1)%2; // switch directions
            calibrStep = 2;
            enable = 0; ECC_controlMove(m,&enable,1); // stop moving
          }
          else
          { // searched both directions and still not valid !
            motors[mot].status &= ~MTR_UNCLBR; // stop the process
            enable = 0; ECC_controlMove(m,&enable,1); // stop moving
            calibrStep = 0;
          }
        }
      }
      else
      { // is valid
        motors[mot].status &= ~MTR_UNCLBR; // stop the process
        motors[mot].status |= MTR_ISCLBR;
        enable = 0; ECC_controlMove(m,&enable,1); // stop moving
        if (ECC_getReferencePosition(m,&refpos) == 0)
        {
          motors[mot].posRef = refpos;
        }
        calibrStep = 0;
        moveto(mot,PROC_STS_START);
      }      
    }

    if (ECC_controlTargetPosition(m,&tgt,0) != 0) continue;
    motors[mot].tgt = tgt;
    if (ECC_getStatusError(m,&sts) == 0)
    {
      if (sts) motors[mot].status = -1; // MST_ERROR ?
    }
    if (motors[mot].stepping)
    {
      if (motors[mot].nsteps > 0) 
      { // step forward
        motors[mot].nsteps--;
        if ((cc=ECC_setSingleStep(m,0)) != 0) 
        {
          motors[mot].status = cc;
          motors[mot].nsteps = 0;
        }
      }
      else if (motors[mot].nsteps < 0)
      { // step backward
        motors[mot].nsteps++;
        if ((cc=ECC_setSingleStep(m,1)) != 0)
        {
          motors[mot].status = cc;
          motors[mot].nsteps = 0;
        }
      }
      if (motors[mot].nsteps == 0) 
      {
        motors[mot].stepping = FALSE;
        motors[mot].status = MTR_IDLE;
      }
    }
    if (ECC_getStatusMoving(m,&moving) == 0)
    {
      if (!moving)
      {
        if ((motors[mot].status & MTR_MOVE) == 0) continue;
        motors[mot].status = MTR_IDLE;
        //motors[mot].pos = motors[mot].tgt;
        motors[mot].moveStart = 0;
        motors[mot].moving = FALSE;
        motors[mot].movpos = FALSE;
        motors[mot].movneg = FALSE;
      }
      else
      {
        if (abs(pos - tgt) < 100)
        {
          enable = 0;
          ECC_controlMove(m,&enable,1);
        }
      }
    }
  }
  return 0;
}
THRHANDLE motorCreateThread(int mot)
{
  DWORD dwThreadId;
  return (THRHANDLE)CreateThread(NULL,0,motorThreadTask,(THREAD_PARAM)mot,0,&dwThreadId);
}
#endif
int initCommMotor(int mot)
{
  int m,cc,retries=0,nmps;
  char *cmd, name[64];
  Bln32 enable = 1, connected = 0;
  Int32 ampl, freq, tgt, refpos;
  ECC_actorType at;

  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  m = motors[mot].axis;

  cmd  = "is connected";
  while (retries++ < 2)
  {
    if ((cc=ECC_getStatusConnected(m,&connected)) == 0) break;
  }
  if (!connected)
  {
    feclog("target motor %d is not connected to the controller!",mot);
    ccerr(device_not_connected);
  }
  ampl = (Int32)motors[mot].voltage * 1000; // send mV to controller
  freq = (Int32)motors[mot].frequency * 1000; // send mHz to controller
  cmd = "enable";
  for (retries=0; retries<2; retries++)
  {
    if ((cc=ECC_controlOutput(m,&enable,1)) == 0) break;
  }
  if (cc) ccerr(cc);
  cmd = "set amplitude";
  for (retries=0; retries<2; retries++)
  {
    if ((cc=ECC_controlAmplitude(m,&ampl,1)) == 0) break;
  }
  if (cc) ccerr(cc);
  cmd = "set frequency";
  for (retries=0; retries<2; retries++)
  {
    if ((cc=ECC_controlFrequency(m,&freq,1)) == 0) break;
  }
  if (cc) ccerr(cc);

  if (ECC_getActorType(m,&at) == 0)
  {
    switch (at)
    {
      case ECC_actorLinear: motors[mot].type = MTR_TYPE_LINEAR; break;
      case ECC_actorGonio: motors[mot].type = MTR_TYPE_GONIOMETER; break;
      case ECC_actorRot:  motors[mot].type = MTR_TYPE_ROTATION; break;
      default: motors[mot].type = MTR_TYPE_UNKNOWN; break;
    }
  }
  if (ECC_getActorName(m,name) == 0)
  {
    strncpy(motors[mot].name,name,32);
  }
  if (ECC_controlTargetPosition(m,&tgt,0) == 0)
  {
    motors[mot].tgt = tgt;
  }
  if (ECC_getReferencePosition(m,&refpos))
  {
    motors[mot].posRef = refpos;
  }
  if (getNmPerStep(mot,&nmps) == 0)
  {
    motors[mot].nmPerStep = (float)nmps;
  }
  motors[mot].enabled = TRUE;
  feclog("attocube successfully initialized");
err:
  if (cc != 0)
  {
    feclog("initialization error: motor %d %s -> <%d>",mot,cmd,cc);
  }
  return cc;
}
void initialize(void)
{
  int i, cc, n, nmotorsFromFile=0;
  char errstr[64], *env=NULL;
 
  if ((cc=csvReadFile(NULL,CONFIG_FILE,&csvCommConfigDb,(void **)&motorInfo)) != 0) 
  {
    feclog("cannot read startup configuration parameters from %s : %s",
      CONFIG_FILE,GetSystemErrorString((short)cc,errstr));
    exit(1);
  }
  nmotorsFromFile = csvCommConfigDb.siz;
 
  if (nmotorsFromFile < 1)
  {
    feclog("no startup configuration parameters from %s given",CONFIG_FILE);
    exit(1);
  }

  if ((cc=ECC_setInit( BIN_PATH )) != 0)
  {
    feclog("could not initialize atto ECC100 : %d",cc);
  }

# if defined(ATTO_TELNET)
  if (strlen(motorInfo[0].target) < 5)
  {
    tineDebug = 1;
    feclog("no valid IP address given for atto Telnet session");
    exit(1);
  }
  tnstartup(motorInfo[0].target,tnport);
# endif
#if defined(ATTO_USB)
  if (strlen(motorInfo[0].target) < 1)
  {
    tineDebug = 1;
    feclog("no valid com port given for atto com session");
    exit(1);
  }
  gIsInProcess = (strnicmp(motorInfo[0].target,"/localhost/cdi",14) == 0) ? TRUE : FALSE;
  feclog("rs232 CDI server is running %s",gIsInProcess ? "in process" : "remotely");
# endif
  gNumMotors = nmotorsFromFile;
  if ((motors=(CommMotor *)calloc(gNumMotors,sizeof(CommMotor))) == NULL)
  {
    feclog("cannot allocate space for %d motors",gNumMotors);
    exit(1);
  }
  if ((mstNames=(NAME64 *)calloc(gNumMotors,sizeof(NAME64))) == NULL)
  {
    feclog("cannot allocate motor name list ! (aborting ...)");
    exit(1);
  }
# ifdef NO_ENCODER
  if ((env=getenv("ATTO_EAGER_SAVERESTORE")) != NULL)
  {
    if (!stricmp(env,"TRUE") || !stricmp(env,"YES"))
    { // this is the default without the environment variable
      eagerSaveRestore = TRUE;
    }
    else
    { // but if the env variable doesn't explicitly set it then ...
      eagerSaveRestore = FALSE;
    }
  }
# endif
  for (i=0; i<gNumMotors; i++)
  {
    if (strlen(motorInfo[i].name) == 0) break;
    strncpy(mstNames[i].name,motorInfo[i].name,32);
    strncpy(motors[i].name,motorInfo[i].name,32);
    strncpy(motors[i].target,motorInfo[0].target,128);
    motors[i].axis = motorInfo[i].axis - 1;  // note the counting offset here with the ECC variety !!
    motors[i].frequency = motorInfo[i].frequency;
    motors[i].voltage = motorInfo[i].voltage;
    for (n=0; n<nmotorsFromFile; n++)
    {
      if (motorInfo[n].motorNr == i+1)
      {
        initCommMotor(i);
        strncpy(motors[i].units,motorInfo[n].units,16);
        motors[i].maxPos = motorInfo[n].posMax;
        motors[i].minPos = motorInfo[n].posMin;
        motors[i].offPos = motorInfo[n].posOff;
        motors[i].scalePos = motorInfo[n].posScale;
        if (motors[i].nmPerStep == 0) motors[i].nmPerStep = (float)motorInfo[n].nmStep;
        motors[i].velMean = (float)(motors[n].nmPerStep * motorInfo[n].frequency);
        if (motors[i].nmPerStep < 1) motors[i].nmPerStep = 1;
        motors[i].type = getMotorType(motorInfo[n].type);
        motors[i].rbTolerance = motorInfo[n].rbTolerance;
        motors[i].velLwr = (float)motorInfo[n].nmStep * FREQ_MIN;
        motors[i].velUpr = (float)motorInfo[n].nmStep * FREQ_MAX;
        motors[i].masterGearRatio = motorInfo[n].gearRatio;
#       ifdef NO_ENCODER
        restoreMotorPosition(i);
#       endif
        break;
      }
    }
    startUpdateMotor(i);
  }
  psAxGrps.idLastScanned = -1; /* unknown at server startup */
}
int SimulateData(int start,int range)
{
  /* This routine is useful for test servers  */
  return start + rand()%(range);
}
void cbPosPsAx(int id, int cc)
{
  psAxGrps.lnkStsPosPsAx = cc;
}
void cbTgtPsAx(int id, int cc)
{
  psAxGrps.lnkStsTgtPsAx = cc;
}
void cbStsPsAx(int id, int cc)
{
  int i;
  psAxGrps.lnkStsStsPsAx = cc;
  if (cc == 0)
  {
    for (i=0; i< psAxGrps.numPsAxGroups; i++)
    { //1:moving,2:idle,3:Error
      if (psAxGrps.stsBuffer[i] == 0) psAxGrps.status[i] = MTR_IDLE;
      if (psAxGrps.stsBuffer[i] == 1) psAxGrps.status[i] = MTR_MOVE;
      if (psAxGrps.stsBuffer[i] == 2) psAxGrps.status[i] = MTR_IDLE;
      if (psAxGrps.stsBuffer[i] == 3) psAxGrps.status[i] = MTR_ERR;
    }
  }
}

void cbV(int id,int cc)
{
  char *c,scratch[64];
  if (cc != 0) return;
  
  if ((c=strchr(motors[id].rbBuf,'=')) == NULL) return;
  strncpy(scratch,++c,64);

  motors[id].voltageRb = (float)strtod(scratch,NULL);

  if (motors[id].voltageRb == 0)  
  {
    motors[id].status &= ~MTR_MOVE;
  }
}
void startUpdateMotor(int mot)
{
  char inpt[64],errstr[64];
  DTYPE dout, din;
  int id, slen, cc=0;

  if (mot < 0 || mot >= gNumMotors) return;

  memset(&dout,0,sizeof(DTYPE)); memset(&din,0,sizeof(DTYPE));
  dout.dFormat = CF_TEXT;
  dout.dArrayLength = 48;
  dout.data.cptr = motors[mot].rbBuf;
  din.dFormat = CF_TEXT;

  sprintf(inpt,"geto %d",motors[mot].axis); slen = strlen(inpt); // to do ???????
  inpt[slen++] = 13; // CR
  inpt[slen++] = 10; // LF
  inpt[slen] = 0;
  din.dArrayLength = slen;
  din.data.cptr = inpt;
# ifndef MULTITHREADED
  if ((id=AttachLinkEx(motors[mot].target,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE,
          MTR_UPDATE_INTERVAL,cbV,CM_TIMER,mot)) < 0) ccerr(-id);
  motors[mot].linkId = id;
# else
  id = -1;
  motors[mot].hIoThread = motorCreateThread(mot);
  if (motors[mot].hIoThread == NULL) ccerr(thread_error);
# endif
err:
  if (cc != 0) 
    feclog("error starting readback loop: %.32s -> %.32s",
      motors[mot].name,GetSystemErrorString((short)cc,errstr));
}
 
LIB_EXPORT int getUnits(int mot,char *unitsBuffer,int bufferLength)
{
  int len;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  len = MIN(bufferLength,16);
  strncpy(unitsBuffer,motors[mot].units,len);
  return 0;
}
LIB_EXPORT int getPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].pos * motors[mot].scalePos + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getGrpPos(int grpId,double *val)
{
  int i, m;
  if (val == NULL) return invalid_parameter;
	for (i=0; i<ngrps; i++)
	{
	  if (grps[i].id != grpId) continue;
	  m = grps[i].grp.grphead - 1;
	  *val = motors[m].pos * motors[m].scalePos + motors[m].offPos;
	  return 0;
	}
  return invalid_index;
}
LIB_EXPORT int getPosMin(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].minPos * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getPosMax(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].maxPos * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getPosOff(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].offPos;
  return 0;
}
LIB_EXPORT int setPosOff(int mot,double val)
{
  int cc;
  char errstr[64];
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  motors[mot].offPos = (float)val;
  // save and restore:
  motorInfo[mot].posOff = motors[mot].offPos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvCommConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getPosScale(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = (double)motors[mot].scalePos;
  return 0;
}
LIB_EXPORT int setPosScale(int mot,double val)
{
  int cc;
  char errstr[64];
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  motors[mot].scalePos = (float)val;
  // save and restore:
  motorInfo[mot].posScale = motors[mot].scalePos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvCommConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getVel(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].velMean;
  return 0;
}
//TODO: trap velocity > upr limit or < lwr limit with software ?
LIB_EXPORT int setVel(int mot,double val)
{
  int cc=0, f, m;
  char dev[64];
  char errstr[64];
 
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (val < motors[mot].velLwr ||
      val > motors[mot].velUpr) return out_of_range;
  sprintf(dev,"%s.Velocity",motors[mot].name);
  if (cc != 0) return cc;
  f = (int)(val/(double)motors[mot].nmPerStep);
  if (f < FREQ_MIN) 
  {
    f = FREQ_MIN;
    val = (double)(motors[mot].nmPerStep / f);
  }
  f *= 1000;
  m = motors[mot].axis;
  if ((cc=ECC_controlFrequency(m,&f,1)) != 0) ccerr(cc);
  motors[mot].frequency = f/1000;
  motors[mot].velMean = (float)(val);

  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvCommConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
err:
  if (cc != 0)
  {
    feclog("could not set velocity! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
LIB_EXPORT int getVelLwr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].velLwr;
  return 0;
}
LIB_EXPORT int getVelUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].velUpr;
  return 0;
}
LIB_EXPORT int getAccUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].accUpr;
  return 0;
}
 
// TODO:
LIB_EXPORT int getAcc(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val =  SimulateData(5,2);
  return 0;
}
//steps per motor unit
LIB_EXPORT int mstepsPerMU(int mot,double *val)
{
  float nmps;
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  nmps = motors[mot].nmPerStep;
  if (nmps == 0) return not_implemented;
  *val = 1.0/nmps; 
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int mstepsPerStep(int mot,int *val)
{
  return not_implemented;
}
//TODO: necessary ?
LIB_EXPORT int stepsPerRadian(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = 1/6.2832;
  return 0;
}
 
int sendMoveCmd(int mot,int useEndSwitch)
{
  int cc,m,steps;
  Bln32 enable = 1;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;

  motors[mot].moving = FALSE;
  motors[mot].movpos = FALSE;
  motors[mot].movneg = FALSE;
  steps = (int)((motors[mot].tgt - motors[mot].pos)/(double)motors[mot].nmPerStep);
  if (steps == 0) return 0;

  motors[mot].moving = TRUE;

  m = motors[mot].axis;
  if ((cc=ECC_controlMove(m,&enable,1)) != 0) ccerr(cc);
err:  
  return cc ? hardware_error : 0;
}
int getNmPerStep(int mot,int *nmps)
{
  int m, i, n, cc = 0, nms1, nms2, tries=1000;
  Int32 pos1, pos2, tgt;
  Bln32 sts, refsts;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  m = motors[mot].axis;
  if ((cc=ECC_getPosition(m,&pos1)) != 0) ccerr(cc);
  ECC_controlTargetPosition(m,&tgt,0);
  ECC_getStatusError(m,&sts);
  ECC_getStatusReference(m,&refsts);
  n = 0;
  while (tries-- > 0 && n < 10)
  {
    if ((cc=ECC_setSingleStep(m,0)) == 0) 
    {
      Sleep(10);
      n++;
    }
  }
  if (n == 0) ccerr(cc);
  for (i=0; i<3; i++)
    if ((cc=ECC_getPosition(m,&pos2)) == 0) break;
  if (cc != 0) ccerr(cc);
  nms1 = (pos2 - pos1)/n;
  n = 0;
  while (tries-- > 0 && n < 10)
  {
    if ((cc=ECC_setSingleStep(m,1)) == 0) 
    {
      Sleep(10);
      n++;
    }
  }
  if (n == 0) ccerr(cc);
  for (i=0; i<3; i++)
    if ((cc=ECC_getPosition(m,&pos1)) == 0) break;
  if (cc != 0) ccerr(cc);
  nms2 = (pos2 - pos1)/n;
  *nmps = (nms1+nms2)/2;
err:
  return cc;
}

int sendStepCmd(int mot,int useEndSwitch)
{
  int steps;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if ((motors[mot].status & (MTR_CLBR|MTR_MOVE)) != 0) return not_allowed;

  motors[mot].moving = FALSE;
  motors[mot].movpos = FALSE;
  motors[mot].movneg = FALSE;
//  steps = (int)((motors[mot].tgt - motors[mot].pos)/(double)motors[mot].nmPerStep);
  steps = motors[mot].nsteps;
  if (steps == 0) steps = 1; // no input -> just step forward

  motors[mot].status |= MTR_STEP;
  motors[mot].stepping = TRUE;
  motors[mot].nsteps = steps;
  return 0;
}
int stopMoveCmd(int mot)
{
  int cc, m;
  Bln32 enable = 0;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;

  m = motors[mot].axis;
  if ((cc=ECC_controlMove(m,&enable,1)) != 0) ccerr(cc);
  motors[mot].moving = FALSE;
  motors[mot].movpos = FALSE;
  motors[mot].movneg = FALSE;
err:
  return cc ? hardware_error : 0;
}
LIB_EXPORT int movests(int mot,int *sts)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (sts == NULL) return invalid_parameter;
  *sts = motors[mot].movCmd;
  return 0;
}
LIB_EXPORT int moveprg(int mot,int *prg)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (prg == NULL) return invalid_parameter;
  *prg = motors[mot].movEtr;
  return 0;
}
LIB_EXPORT int moveto(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      motors[mot].moveStart = 0;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendMoveCmd(mot,TRUE) == 0)
      {
        motors[mot].moveStart = makeDataTimeStamp();
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      motors[mot].moveStart = makeDataTimeStamp();
      break;
    case PROC_STS_RESUME:
      if (sendMoveCmd(mot,TRUE) == 0)
      {
        motors[mot].moveStart = makeDataTimeStamp();
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      motors[mot].moveStart = 0;
      break;
  }
  return 0;
}
LIB_EXPORT int stepto(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      motors[mot].moveStart = 0;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendStepCmd(mot,TRUE) == 0)
      {
        motors[mot].moveStart = makeDataTimeStamp();
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      motors[mot].moveStart = makeDataTimeStamp();
      break;
    case PROC_STS_RESUME:
      if (sendStepCmd(mot,TRUE) == 0)
      {
        motors[mot].moveStart = makeDataTimeStamp();
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      motors[mot].moveStart = 0;
      break;
  }
  return 0;
}
//TODO: test ?
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendMoveCmd(mot,FALSE) == 0)
      {
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    default:
      return moveto(mot,cmd);
  }
  return 0;
}
int getGrpTblIndex(MotorGroup *grp) 
{ 
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (memcmp(&grps[i].grp,grp,sizeof(MotorGroup))) continue;
    return i;
  }
  return -1; 
}
int addToGrpTbl(MotorGroup *grp) 
{ 
  int i = getGrpTblIndex(grp);
  if (i >= 0) return i;
  if (ngrps < GRPTBL_SIZE)
  {
    i = ngrps;
    grps[i].gearRatio = 0;
    memcpy(&grps[i].grp,grp,sizeof(MotorGroup));
    ngrps++;
    return i;
  }
  return -1; 
}
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp)
{
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    memcpy(grp,&grps[i].grp,sizeof(MotorGroup));
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int setGrpMembers(int grpId,MotorGroup *grp)
{
  int g;
  if (grpId < MST_DYNAGRP_OFFST)
  {
    if ((g=getPsAxGroupIndex(grpId)) < 0)
    {
      if (psAxGrps.numPsAxGroups >= PSAX_BUFFER_SIZE) return resources_exhausted;
      g = psAxGrps.numPsAxGroups;
      psAxGrps.numPsAxGroups++;
    }
    psAxGrps.grpIdXref[g] = grpId;
  }
  else
  {
    if ((g=addToGrpTbl(grp)) < 0) return invalid_index;
    grps[g].id = grpId;
  }
//  memcpy(&grps[g].grp,grp,sizeof(MotorGroup));
  return 0;
}
LIB_EXPORT int setGrpGearRatio(int grpId,double gearRatio)
{
  int i;
  if (gearRatio == 0) return out_of_range;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    grps[i].gearRatio = (float)gearRatio;
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int getGrpGearRatio(int grpId,double *gearRatio)
{
  int i;
  if (gearRatio == NULL) return invalid_parameter;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    *gearRatio = (double)grps[i].gearRatio;
    return 0;
  }
  return invalid_index;
}
//TODO: fix
LIB_EXPORT int grpmoveto(int grpId,int cmd)
{
  int cc=0, i, g=-1;
  MotorGroup grp;
  double etr, gearRatio=0;
  float fvals[4];

  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  if (grp.grphead > 0) 
  { // group master defined from above
    g = grp.grphead - 1;
  }
  else
  { // not defined: take the first in the list
    for (i=0; i<gNumMotors; i++) if (IsMotorInGroup(i,&grp)) break;
    if (i == gNumMotors) return configuration_error;
    g = i;
  }
  if (cmd == PROC_STS_START)
  { // first things first
    for (i=0; i<gNumMotors; i++)
    { 
      if (IsMotorInGroup(i,&grp))
      { // don't allow if anybody is being calibrated or is moving !
        if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
        if ((motors[i].status & MTR_MOVE) != 0) return operation_busy;
        if (g == i) 
        { // fix the coupling parameters 
          getGrpGearRatio(grpId,&gearRatio);
          if (gearRatio == 0) gearRatio = (double)motors[i].masterGearRatio;
          fvals[0] = 1;
          fvals[1] = (float)(g + 1);
          fvals[2] = (float)gearRatio;
        }
      }
    }
    for (i=0; i<gNumMotors; i++)
    { // couple all motors in group to same axis
      if (IsMotorInGroup(i,&grp))
      { // this one is a group member !      
        motors[i].grpId = grpId; // make a note of the group id
        if (g == i) continue; // don't couple the master
//        sprintf(dev,"%s.CoupleAxis",motors[i].name);
//        if ((cc=sendFloatArrayToHardware(dev,fvals,3)) != 0)
        { // couple all other members to the master
          motors[i].status |= MTR_ERR;
          continue;
        }
      }
    }
  }
  for (i=0; i<gNumMotors; i++)
  { // send move commands only to the master
    if (IsMotorInGroup(i,&grp))
    {       
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          motors[i].grpId = -1;
          break;
        case PROC_STS_START:
          if (i == g) moveto(g,cmd);
          etr = fabs(motors[i].tgt - motors[i].pos);
          etr /= (double)(motors[i].velMean);
          //etr /= (double)(motors[i].velRps * motors[i].mstepsPerCycle);
          motors[i].movEtr = (int)etr;
          motors[i].movPaused = FALSE;
          motors[i].status |= MTR_MOVE;
          break;
        case PROC_STS_PAUSE:
          if (i == g) stopMoveCmd(g);
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_RESUME:
          if (i == g) moveto(g,cmd);
          motors[i].movPaused = FALSE;
          motors[i].status &= ~MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          if (i == g) stopMoveCmd(g);
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
      }
    }
  }
  if (cmd == PROC_STS_STOP)
  { // finished
    for (i=0; i<gNumMotors; i++)
    { // decouple all coupled motors from axis
      if (IsMotorInGroup(i,&grp))
      {       
        if (i == g) continue;
//        sprintf(dev,"%s.DecoupleAxis",motors[i].name);
//        if ((cc=sendFloatToHardware(dev,1)) != 0)
        {
          motors[i].status |= MTR_ERR;
          continue;
        }
      }
    }
  }
  return cc;
}
LIB_EXPORT int setPos(int mot,double val)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (motors[mot].maxPos > motors[mot].minPos)
  { // make sure these have been supplied in the configuration before making this check
    if (val < motors[mot].minPos || val > motors[mot].maxPos) 
    {
      if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
    }
  }
  motors[mot].pos = val;
# ifdef NO_ENCODER
  saveMotorPostion(mot);
# endif
  return 0;
}
LIB_EXPORT int setTgtPos(int mot,double val)
{
  int cc = 0, m, tgt;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (val < motors[mot].minPos || val > motors[mot].maxPos) 
  {
    if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
  }
  tgt = (int)val;
  m = motors[mot].axis;
  if ((cc=ECC_controlTargetPosition(m,&tgt,1)) == 0)
  {
    motors[mot].tgt = tgt;  
  }
  return cc;
}
LIB_EXPORT setTgtSteps(int mot,double val)
{
  int cc = 0;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  motors[mot].nsteps = (int)val;
  return cc;
}

LIB_EXPORT int getTgtPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
 
  *val = motors[mot].tgt * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int setGrpTgtPos(int grpId,double val)
{
  int cc, i, g=-1;
  MotorGroup grp;

  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors; i++)
  { // don't allow if anybody is being calibrated !
    if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
  }
  for (i=0; i<gNumMotors && g < 0; i++) if (IsMotorInGroup(i,&grp)) g = i;
  return setTgtPos(g,val);
}
LIB_EXPORT int getGrpTgtPos(int grpId,double *val)
{
  int cc, i, g=-1;
  MotorGroup grp;

  if (val == NULL) return invalid_parameter;
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors && g < 0; i++) if (IsMotorInGroup(i,&grp)) g = i;
  if (g < 0 || g > gNumMotors) return illegal_device;
  return getTgtPos(g,val);
}
//TODO GROUP ???
LIB_EXPORT int calibr(int mot,int *sts)
{
  int cc = 0;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if ((motors[mot].status & 0xff) != MTR_IDLE) return not_allowed;
  motors[mot].pos = 0;
  motors[mot].tgt = 0;
# ifdef NO_ENCODER
  saveMotorPostion(mot);
# endif
  motors[mot].status = MTR_UNCLBR;
  calibrStep = 0;
  cc = ECC_setReset(motors[mot].axis);

  return cc;
}
//TODO: check
LIB_EXPORT int calibrToOffset(int mot,double val)
{
  int cc = 0;

  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (motors[mot].type != MTR_TYPE_LINEAR) return not_allowed;
  motors[mot].status = MTR_UNCLBR;
  motors[mot].pos = 0;
  motors[mot].tgt = (float)val;
//  moveto(mot,PROC_STS_START);
//  sprintf(dev,"%.32s.Home",motors[mot].name);
//  cc = motors[mot].cmdSts = sendDoubleToHardware(dev,val);
  return cc;
}
LIB_EXPORT int getonline(int mot,int *sts)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device; 
  if (sts == NULL) return argument_list_error;
  *sts = motors[mot].enabled;
  return 0;
}
LIB_EXPORT int setonline(int mot,int sts)
{
  int cc = 0;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  motors[mot].enabled = sts;
  // to do ???????
  return 0;
}

LIB_EXPORT int reset(int mot,int *sts)
{
  return calibr(mot,sts);
}
//TODO: necessary ?
int chkEndStation(int mot,int mask)
{
//char dev[128];
  int reg = 0;
//sprintf(dev,"%s.regExSta",motors[mot].reg);
//getIntegerFromHardware(dev,&reg);
  return (reg & mask) ? TRUE : FALSE;
}
LIB_EXPORT int getSts(int mot,int *sts)
{ // is called from the background task @ 10 Hz
  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *sts = motors[mot].status;
  return 0;
}
LIB_EXPORT int getGrpSts(int grpId,int *sts)
{
  int i, m;
  if (sts == NULL) return invalid_parameter;
    for (i=0; i<ngrps; i++)
    {
      if (grps[i].id != grpId) continue;
      m = grps[i].grp.grphead - 1;
      *sts = motors[m].status;
      return 0;
    }
  return invalid_index;
}
int makeMotStsBits(int mot)
{
  int stsBits=0;
  if (mot < 0 || mot >= gNumMotors) return 0;
  if (motors[mot].idle) stsBits |= STS_IDLE;
  if (motors[mot].enabled) stsBits |= STS_ENABLED;
  if (motors[mot].moving) stsBits |= STS_MOVING;
  if (motors[mot].movpos) stsBits |= STS_MOVPOS;
  if (motors[mot].movneg) stsBits |= STS_MOVNEG;
  if (motors[mot].endpos) stsBits |= STS_ENDPOS;
  if (motors[mot].endneg) stsBits |= STS_ENDNEG;
  return stsBits;
}
LIB_EXPORT int getStsInfo(int mot,MotorStatus *motsts)
{
  if (motsts == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  strncpy(motsts->motName,motors[mot].name,64);
  motsts->stsBits = makeMotStsBits(mot);
  motsts->stsAxis = motors[mot].axisState;
  motsts->stsCoupled = motors[mot].coupledState;
  motsts->stsClbr = motors[mot].clbrSts;
  motsts->stateClbr = motors[mot].clbrState;
  motsts->idxCntlLoop = motors[mot].actCntrlLoopIndex;
  motsts->pos = (float)motors[mot].posRb;
  motsts->tgt = (float)motors[mot].tgtRb;
  motsts->lag = motors[mot].lag;
  motsts->vel = motors[mot].velRb;
  motsts->load = motors[mot].pcntMaxCur;
  motsts->ovrRide = motors[mot].overrideSts;
  return 0;
}

LIB_EXPORT int getLastHdwErr(int mot,NAME64 *errstr,int *errcode)
{
  char dev[128];
  int cc = 0;
  if (errstr == NULL && errcode == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (errstr != NULL) 
  { 
    sprintf(dev,"%.32s.ErrorText",motors[mot].name);
//   cc = getCharArrayFromHardware(dev,errstr->name,64);
    if (cc != 0) return cc;
  }
  if (errcode != NULL)
  { 
    sprintf(dev,"%.32s.ErrorId",motors[mot].name);
//   cc = getIntegerFromHardware(dev,errcode);
    if (cc != 0) return cc;
  }
  return 0;
}

XMA xprps[] = { 
  {"Amplitude", "stepping amplitude", "V", 10000, 0, CA_READ|CA_WRITE},
  {"Frequency", "stepping frequency", "Hz", 256, 0, CA_READ|CA_WRITE},
  {"Voltage", "readback output voltage", "V", 256, 0, CA_READ},
  {"NanoMetersStep","find nanometers per step", "nm/step", 1, 0, CA_READ},
  {"ReferencePosition", "reference position", "nm", 1, 0, CA_READ},
};
int nxprps = sizeof(xprps)/sizeof(XMA);
int getMotorFromName(char *name)
{
  int i;
  for (i=0; i<gNumMotors; i++) if (!strnicmp(mstNames[i].name,name,32)) return i;
  return -1;
}
int attoeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int mot=getMotorFromName(devName);
  int cc,iv;
  float fv;
  char scratch[64];
  // devName is one of the mstNames
  // devProperty is one of the xprps
  // dout probably asks for all or 1 double value
  // din will only set 1 double value
  // access will specify CA_WRITE or not
  if (mot < 0) return illegal_device;

  if (stricmp(devProperty,"NanoMetersStep") == 0)
  {
    int nmps = 0;
    cc = getNmPerStep(mot,&nmps);
    return PutValuesFromLong(dout,&nmps,1);
  }
  if (stricmp(devProperty,"ReferencePosition") == 0)
  {
    return PutValuesFromDouble(dout,&motors[mot].posRef,1);
  }
  if (stricmp(devProperty,"Amplitude") == 0 || stricmp(devProperty,"Voltage") == 0)
  {
    if (access & CA_WRITE)
    {
      if (din->dArrayLength < 1) return dimension_error;
      if ((cc=GetValuesAsFloat(din,&fv,1)) != 0) return cc;
      if (fv < VOLT_MIN || fv > VOLT_MAX) return out_of_range;
      iv = (int)fv;
      if ((cc=ECC_controlAmplitude(motors[mot].axis,&iv,1)) != 0) return hardware_error;
      fv *= 1000;
      motors[mot].voltage = fv;
      fv /= 1000;
      if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvCommConfigDb)) != 0) 
      {
        feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,scratch));
      }
    }
    return PutValuesFromFloatEx(dout,&motors[mot].voltage,1,0);
  }
  if (stricmp(devProperty,"Frequency") == 0)
  {
    if (access & CA_WRITE)
    {
      if (din->dArrayLength < 1) return dimension_error;
      if ((cc=GetValuesAsLong(din,&iv,1)) != 0) return cc;
      if (iv < FREQ_MIN || iv > FREQ_MAX) return out_of_range;
      iv *= 1000;
      if ((cc=ECC_controlFrequency(motors[mot].axis,&iv,1)) != 0) return hardware_error;
      iv /= 1000;
      motors[mot].frequency = iv;
      motors[mot].velMean = (float)(motors[mot].nmPerStep * motors[mot].frequency);
      if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvCommConfigDb)) != 0) 
      {
        feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,scratch));
      }
    }
    return PutValuesFromLongEx(dout,&motors[mot].frequency,1,0);
  }

  return not_implemented;
}
int startup(void)
{
  initialize();
  SetMotorList(mstNames,gNumMotors); // done
  SetXMAList(xprps,nxprps,attoeqm);
  SetStatusList(stsLst,sizStsLst);   // done 
  SetGetMotorStatusDispatch(getSts); // done
  SetGetGroupMotorStatusDispatch(getGrpSts);
  SetGetMotorVelocityDispatch(getVel); // done
  SetSetMotorVelocityDispatch(setVel); // done
  SetGetMotorAccelerationDispatch(getAcc);
  SetGetMotorPositionDispatch(getPos); // done
  SetGetGroupMotorPositionDispatch(getGrpPos);
  SetSetMotorPositionDispatch(setPos);
  SetGetMotorStepsPerMUDispatch(mstepsPerMU); // done
  SetGetMotorMicroStepsPerStepDispatch(mstepsPerStep); // done
  SetGetMotorStepsPerRadianDispatch(stepsPerRadian); // done
  SetGetMoveToStatusDispatch(movests); // done
  SetGetMoveToProgressDispatch(moveprg); // done
  SetGetGroupMoveStatusDispatch(movests);
  SetGetIncrementMoveStatusDispatch(movests);
  SetGetIncrementGroupMoveStatusDispatch(movests);
  SetSetMoveToStatusDispatch(moveto); // done
  SetSetStepToStatusDispatch(stepto);
  SetSetFreeMoveToStatusDispatch(movetoNoEndSwitch); // done
  SetSetGroupMoveStatusDispatch(grpmoveto); // done ?
  SetSetIncrementGroupMoveStatusDispatch(grpmoveto);
  SetSetIncrementMoveStatusDispatch(moveto); // done 
  SetSetMotorTargetPositionDispatch(setTgtPos); // done
  SetGetMotorTargetPositionDispatch(getTgtPos); // done
  SetSetMotorTargetStepsDispatch(setTgtSteps);

 // SetSetGroupMotorTargetPositionDispatch(setGrpTgtPos);
 // SetGetGroupMotorTargetPositionDispatch(getGrpTgtPos);
  SetGetGroupMembersDispatch(getGrpMembers); // done ?
  SetSetGroupMembersDispatch(setGrpMembers); // done ?
  SetCalibrateMotorDispatch(calibr);  // done
  SetResetMotorDispatch(reset); // done
  SetGetMotorVelocityLowerLimitDispatch(getVelLwr);
  SetGetMotorVelocityUpperLimitDispatch(getVelUpr);
  SetGetMotorAccelerationUpperLimitDispatch(getAccUpr);
  SetGetMotorPositionMinimumDispatch(getPosMin); // done
  SetGetMotorPositionMaximumDispatch(getPosMax); // done
  SetGetMotorPositionOffsetDispatch(getPosOff); // done
  SetSetMotorPositionOffsetDispatch(setPosOff); // done
  SetGetMotorPositionScaleDispatch(getPosScale); // done
  SetSetMotorPositionScaleDispatch(setPosScale); // done
  SetCalibrateToOffsetDispatch(calibrToOffset);
  SetGetMotorUnitsDispatch(getUnits);
  SetGetMotorOnlineDispatch(getonline);
  SetSetMotorOnlineDispatch(setonline);
  SetGetMotorStatusBlockDispatch(getStsInfo);
  SetSetGroupGearRatioDispatch(setGrpGearRatio); // done
  SetGetGroupGearRatioDispatch(getGrpGearRatio); // done
  SetGetLastHardwareErrorDispatch(getLastHdwErr);
  // allow access to a debug flag
  RegisterUserCommand("attodebug",NULL,(int *)&attoDebugLevel,NULL,CA_READ|CA_WRITE);
# ifdef NO_ENCODER
  RegisterUserCommand("eagersave",NULL,(int *)&eagerSaveRestore,NULL,CA_READ|CA_WRITE);
# endif
//  RegisterUserCommand("mode",getsetm,NULL,NULL,CA_READ|CA_WRITE);
  return 0;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
#     if defined(ATTO_USB)
      LoadLibrary("cdi32.dll");
#     endif
      SetInitMotorHardwareDispatch(startup);
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      mstCommTerminating = TRUE;
#     if defined(ATTO_TELNET)
      soclose(m_hSocket);
#     endif
#     ifdef NO_ENCODER
      saveMotorPostion(0);
#     endif
      break;
  }
  return (TRUE);
}
 