#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#define MULTITHREADED
#include "tine.h"
#include "hvpositionerv2.h"
#include "mstAtto.h"
 
#define USE_ABSOLUTESYNC
 
#define CLBR_PATTERN 32641
#define MOVETO_DURATION 600

GETGMTPFCNP getGroupMemberTargetPositionFcn = NULL;
GETGPFCNP getGroupPositionFcn = NULL;

MXHANDLE hAttoMutex;
MXHANDLE hCalbrMutex;
#ifdef WIN32
// unexported routines from tine32.lib
#define ReleaseSystemMutex(m) ReleaseMutex(m)
#define CloseSystemMutex(m) CloseHandle(m)
int InitMutex(MXHANDLE *mx)
{
  if (mx == NULL) return -1;
  *mx = CreateMutex(NULL,FALSE,NULL);
  if (*mx == NULL) return -2;
  return 0;
}
int WaitForMutex(MXHANDLE mx,int to)
{
  int cc;
  if (mx == (MXHANDLE)0) return 0;
  cc = WaitForSingleObject(mx,(to) < 0 ? INFINITE : to);
  if (cc == WAIT_FAILED) cc = GetLastError();
  return cc == WAIT_OBJECT_0 ? 0 : cc;
}
#endif
NAME32 stsLst[] = {
  "moving",
  "paused",
  "",
  "",
  "calibrating (step1)",
  "calibrating (step2)",
  "calibrating (step3)",
  "",
  "",
  "",
  "",
  "is calibrating",
  "uncalibrated",
  "",
  "",
  "hardware error",
};
 
int sizStsLst = sizeof(stsLst)/sizeof(NAME32);
 
int attoDebugLevel = 0;
 
int getMotorType(char *typeStr)
{
  if (typeStr == NULL) return MTR_TYPE_UNKNOWN;
  if (!stricmp(typeStr,"LINEAR")) return MTR_TYPE_LINEAR;
  if (!stricmp(typeStr,"ROTATION")) return MTR_TYPE_ROTATION;
 return MTR_TYPE_UNKNOWN;
}
 
#define STS_BUFFER_SIZE 20
#define PSAX_BUFFER_SIZE 20

typedef struct
{
  double rbBuffer[PSAX_BUFFER_SIZE];
  double tgtBuffer[PSAX_BUFFER_SIZE];
  int stsBuffer[PSAX_BUFFER_SIZE];
  int grpIdXref[PSAX_BUFFER_SIZE];
  int status[PSAX_BUFFER_SIZE];
  int lnkStsPosPsAx; 
  int lnkIdPosPsAx; 
  int lnkStsStsPsAx; 
  int lnkIdStsPsAx;
  int lnkIdTgtPsAx;
  int lnkStsTgtPsAx;
  int numPsAxGroups;
  int idLastScanned;
} PseudoAxisGroups;

PseudoAxisGroups psAxGrps;

typedef struct
{
  int id;
  float gearRatio;
  MotorGroup grp;
} MotGrpTbl;
#define GRPTBL_SIZE 10
MotGrpTbl grps[GRPTBL_SIZE];
int ngrps = 0;

// status bit definitions (property "StatusInfo" -> stsBits)
#define STS_IDLE 0x01
#define STS_ENABLED 0x02
#define STS_MOVING 0x04
#define STS_MOVPOS 0x08
#define STS_MOVNEG 0x10
#define STS_ENDPOS 0x20
#define STS_ENDNEG 0x40
#define STS_OVERRIDE 0x80

typedef struct
{
  char name[32];
  char cfg[64];
  char units[16];
  int type;
  int motid;
  int grpId;   // -1 if not currently a group member
  int idle;    // bit
  int enabled; // bit
  int moving;  // bit
  int stepping;
  int movpos;  // bit
  int movneg;  // bit
  int endpos;  // bit
  int endneg;  // bit
  int axisState; // word ? byte ?
  int coupledState;// word ? byte ?
  int clbrState;// current calibration state in calibration procedure
  int clbrSts;  // is calibrated or not
  int actCntrlLoopIndex;// word ? byte ?
  int microStepsPerStep;
  int mstepsPerCycle;
  float cyclesPerMM;
  float mstepsPerMM;
  float stsBuffer[STS_BUFFER_SIZE];
  double pos;
  double posRb; // readback position
  double tgt;
  double tgtRb; // readback target position
  float overrideSts;
  float lag;  // soll - ist
  float velRb;
  float pcntMaxCur;
  float maxPos; // maximum position (micro steps)
  float minPos; // minimum position (micro steps)
  float offPos; // offset position (micro steps)
  float scalePos; // scale factor (usually -1.0 or 1.0)
  float velMean; // normal velocity (msteps / sec)
  float velMin;
  float velMax;
  float velLwr;
  float velUpr;
  float accUpr;
  float rbTolerance;
  float masterGearRatio; // 1.0 default
  int movCmd;
  int movEtr; // 'estimated time remaining' for move action
  int movPaused;
  int stsRb;
  int status;
  int cmdSts;
  THRHANDLE hIoThread;
  int hPositioner;
  int positioner;
  int axis;
  int amplitude;
  int frequency;
  int rotCount;
  int valid;
  time_t lastCheckedClbr;
} AttoMotor;
 
AttoMotor *motors = NULL;
 
int gNumMotors = 1;
int gIsInProcess = FALSE;
void cbTgtPos(int id, int cc);
void startUpdateMotor(int mot);
LIB_EXPORT int moveto(int mot,int cmd);
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd);
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp);
LIB_EXPORT int grpmoveto(int grpId,int cmd);
LIB_EXPORT int incgrpmoveto(int grpId,int cmd);
int stopMoveCmd(int mot);
 
typedef struct
{
  char lib[64];
  char name[64];
  char cfg[64];
  char type[32];
  char units[16];
  int motorNr;
  int positioner;
  int axis;
  int mstepsStep;
  int mstepsCycle;
  float velNormal;
  float velMin;
  float velMax;
  float posMax;
  float posMin;
  float posOff;
  float posScale;
  float mstepsMM;
  float rbTolerance;
  float gearRatio;
} MotorInfoStruct;
 
CSVCOLUMN csvAttoConfigColumns[] =
{
  { "LIBRARY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(MotorInfoStruct,lib),NULL,0 },
  { "NAME","",0,CF_TEXT,64,(BYTE *)OFFSETIN(MotorInfoStruct,name),NULL,0 },
  { "CONFIG","Ant200.aps",0,CF_TEXT,64,(BYTE *)OFFSETIN(MotorInfoStruct,cfg),NULL,0 },
  { "TYPE","LINEAR",0,CF_TEXT,32,(BYTE *)OFFSETIN(MotorInfoStruct,type),NULL,0 },
  { "MOTOR","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,motorNr),NULL,0},
  { "POSITIONER","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,positioner),NULL,0 },
  { "AXIS","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,axis),NULL,0 },
  { "UNITS","nm",0,CF_TEXT,16,(BYTE *)OFFSETIN(MotorInfoStruct,units),NULL,0},
  { "MSTEPS_MM","12800",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsMM),NULL,0 },
  { "MSTEPS_CYCLE","200",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsCycle),NULL,0 },
  { "MSTEPS_STEP","64",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsStep),NULL,0 },
  { "POSITION_MAXIMUM","768000",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMax),NULL,0 },
  { "POSITION_MINIMUM","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMin),NULL,0 },
  { "POSITION_OFFSET","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posOff),NULL,0 },
  { "POSITION_SCALE","1",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posScale),NULL,0 },
  { "VELOCITY_NORMAL","0.1",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,velNormal),NULL,0 },
  { "VELOCITY_MINIMUM","0.01",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,velMin),NULL,0 },
  { "VELOCITY_MAXIMUM","0.5",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,velMax),NULL,0 },
  { "POSITION_TOLERANCE","0.01",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,rbTolerance),NULL,0 },
  { "GEAR_RATIO","1.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,gearRatio),NULL,0 },
};
CSVDB csvAttoConfigDb =
{
  csvAttoConfigColumns,
  sizeof(csvAttoConfigColumns)/sizeof(csvAttoConfigColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(MotorInfoStruct)
};

typedef struct
{
  int number;
  int deviceCount;
  int handle;
} AttoPositioner;

AttoPositioner attoPositioner[32];
int nAttoPositioners = 0;

int ncb2cc(int ncb_err)
{
  switch (ncb_err)
  {
    case NCB_Ok:/**< Kein Fehler                        */
      return operation_success; // '0' in both cases
    case NCB_Timeout: /**< Timeout bei Datenabruf             */
      return operation_timeout;
    case NCB_NotConnected:/**< Kein Kontakt zum Positioner ber USB    */
      return device_not_connected;
    case NCB_DriverError: /**< Fehler bei der Treiberansprache    */
      return hardware_error;
    case NCB_BootIgnored: /**< Booten ignoriert, Gert lief schon */
      return not_accepted;
    case NCB_FileNotFound:/**< Boot-Image nicht gefunden          */
      return no_such_file;
    case NCB_InvalidParam: /**< bergebener Parameter ungltig     */
      return invalid_parameter;
    case NCB_DeviceLocked: /**< Ein Verbindungsversuch schlug fehl, da das Device schon verwendet wird */
      return access_locked;
    case NCB_NotSpecifiedParam: /**< bergebener Parameter ist auerhalb der Spezifikation     */
      return out_of_range;
    case NCB_Error: /**< Unbekannter/sonstiger Fehler       */
    default:
      break;
  }
  return hardware_error;
}
int getPositionerHandle(int index)
{
  int i=0,n,ncb_cc,cc,hndl=0,id;
  PositionerInfo *pi = 0;
  char errstr[64];
  static int scan_done = FALSE;

  if (!scan_done)
  {
    if (WaitForSystemMutex(hAttoMutex,-1) != 0) return -semaphore_busy;
    if ((n=PositionerCheck(&pi)) < 1) 
    {
      feclog("no positioners found !");
      return -hardware_error;
    }
    for (id=0; id<n; id++)
    { // run thru what was found 
      for (i=0; i<nAttoPositioners; i++)
      { // cross-check against what's been scanned
        if (attoPositioner[i].number == pi[id].id) break;
      }
      attoPositioner[i].deviceCount++;
      if (i < nAttoPositioners) continue;  
      attoPositioner[i].number = pi[id].id;
      ncb_cc = PositionerConnect(id,&hndl);
      cc = ncb2cc(ncb_cc);
      if (cc != 0) 
      { // couldn't connect !
        feclog("cannot connect to positioner %d : <%d> %.32s",
          pi[id].id,ncb_cc,GetSystemErrorString((short)cc,errstr));
        continue;
      }
      attoPositioner[i].deviceCount = 1;  //TODO: how to deterime this ? (active axes?)
      attoPositioner[i].handle = hndl;
      millisleep(100); // necessary ?
  //  PositionerStaticAmplitude(hndl,30); //TODO:  static amplitude = 30 is hard-coded here !!!!!!
      nAttoPositioners++;    
      if (nAttoPositioners > 32) 
      {
        feclog("more than 32 positioners found !");
        return -resources_exhausted;
      }
    }
    ReleaseSystemMutex(hAttoMutex);
    scan_done = TRUE;
    millisleep(500); // necessary ?
  }

  for (i=0; i<nAttoPositioners; i++)
  {
    if (attoPositioner[i].number == index)
      return attoPositioner[i].handle;
  }
  feclog("desired positioner id %d not found !\n",index);
  return -1;
}

int attoShutdown(void)
{
  int i;
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
  for (i=0; i<nAttoPositioners; i++) PositionerClose(attoPositioner[i].handle);
  ReleaseSystemMutex(hAttoMutex);
  return 0;
}

NAME64 *mstNames = NULL;
 
int getPsAxGroupIndex(int grpId)
{
  int i;
  for (i=0; i<psAxGrps.numPsAxGroups; i++)
  {
    if (psAxGrps.grpIdXref[i] == grpId) return i;
  }
  return -1;
}
static MotorInfoStruct *motorInfo = NULL;

#define ATTO_FUDGE_DISTANCE 100000
#define MTR_UPDATE_INTERVAL 100
#define MTR_ZERO_TOLERANCE 500
int mstAttoTerminating = FALSE;
int *calibrStep = NULL;
int calibratingMotor = -1;
THREAD_TYPE motorThreadTask(THREAD_PARAM lpvThreadParam)
{
  int mot = (int)lpvThreadParam;
  int i, pos=0, vel, sts=0, cc, isMoving=FALSE, stsMoving, stsStable;
  int hnd = motors[mot].hPositioner;
  int pass  = 0;
  double dpos;
  static int *lastStsMoving = NULL;
  static int *calibrDir = NULL; // 1 = forward, -1 = backward
  static int *notmovingCounter = NULL;
  static int done = FALSE;
  static int counts = 0;

  if (!done)
  {
    done = TRUE;
    notmovingCounter = (int *)calloc(gNumMotors,sizeof(int));
    if (notmovingCounter == NULL) printf("out of memory !!!\n"), exit(1);
    lastStsMoving = (int *)calloc(gNumMotors,sizeof(int));
    if (lastStsMoving == NULL) printf("out of memory !!!\n"), exit(1);
    calibrDir = (int *)calloc(gNumMotors,sizeof(int));
    if (calibrDir == NULL) printf("out of memory !!!\n"), exit(1);
    for (i=0; i<gNumMotors; i++) calibrDir[i] = 1;
  }

  // can't read the target setting, so take the position
  // readback at server start:
  if (WaitForSystemMutex(hAttoMutex,-1) == 0)
  {
    cc = PositionerGetPosition(hnd,motors[mot].axis,&pos);
    motors[mot].tgt = (double)pos;
    ReleaseSystemMutex(hAttoMutex);
  }
  while (!mstAttoTerminating)
  {
    if (WaitForSystemMutex(hAttoMutex,-1) != 0) goto bow;
    cc = PositionerGetSpeed(hnd,motors[mot].axis,&vel);
    motors[mot].velMean = (float)vel; 
    cc = PositionerGetStatus(hnd,motors[mot].axis,&sts);
    motors[mot].stsRb = sts;
    cc = PositionerGetPosition(hnd,motors[mot].axis,&pos);
    dpos = (double)pos;
    ReleaseSystemMutex(hAttoMutex);
    stsMoving = (motors[mot].stsRb & 0x01);
    stsStable = (lastStsMoving[mot] == stsMoving) ? TRUE : FALSE;
    if (counts++ < gNumMotors) stsStable = FALSE;
    lastStsMoving[mot] = stsMoving;
    if (stsStable)
    { // software check on 'is moving' ...
      isMoving = (fabs(dpos - motors[mot].pos) > MTR_ZERO_TOLERANCE) ? TRUE : FALSE;
    }
    if (stsMoving && !isMoving) notmovingCounter[mot]++;
    if (notmovingCounter[mot] > 2)
    {
      notmovingCounter[mot] = 0;
//      stopMoveCmd(mot); // officially stop moving
    }
    motors[mot].pos = dpos;
    if (motors[mot].status & MTR_UNCLBR)
    { // calibration signal ...
      bool valid = false;
      int rpos, tpos;
      if (calibratingMotor < 0) calibratingMotor = mot;
      if (calibratingMotor != mot) goto bow;
      Sleep(100);
      if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
      if ((cc=PositionerGetReference(motors[mot].hPositioner,motors[mot].axis,&rpos,&valid)) != 0) ccerr(cc);
      dbglog("%.32s calibration step %d, moving %s, valid %s",
        motors[mot].name,calibrStep[mot],(isMoving ? "TRUE" : "FALSE"),(valid ? "TRUE" : "FALSE"));
	    motors[mot].valid = valid ? TRUE : FALSE;
      ReleaseSystemMutex(hAttoMutex);
      Sleep(1000);
      if (!valid)
      {
        if (!isMoving)
        { // not valid and not moving
          if (calibrStep[mot] > 10)
          { // everything looked okay, but not we're reading not valid again ?
            if (WaitForSystemMutex(hAttoMutex,-1) != 0) goto bow;
            for (i=-1; i<2; i += 2)
            {
              dbglog("%.32s calibration: finished but read valid = false (try a nudge)",motors[mot].name);
              PositionerGetPosition(hnd,motors[mot].axis,&pos);
              tpos = (int)(i*ATTO_FUDGE_DISTANCE);
              if ((cc=PositionerMoveAbsolute(motors[mot].hPositioner,motors[mot].axis,tpos,motors[mot].rotCount)) != 0) ccerr(cc);
              isMoving = TRUE;
              while (isMoving)
              {
                dpos = (double)pos;
                Sleep(100);
                PositionerGetPosition(hnd,motors[mot].axis,&pos);
                motors[mot].pos = (double)pos;
                isMoving = (fabs(dpos - motors[mot].pos) > MTR_ZERO_TOLERANCE) ? TRUE : FALSE;
              }
            }
            ReleaseSystemMutex(hAttoMutex);
            dbglog("%.32s calibration: finished declared",motors[mot].name);
            goto err; // best we can do ...
          }
          else if ((calibrStep[mot]%2) == 0)
          { // starting off
            if (WaitForSystemMutex(hAttoMutex,-1) != 0) goto bow;
            tpos = (int)(calibrDir[mot]*motors[mot].maxPos);
            dbglog("%.32s calibration: move to %d",motors[mot].name,tpos);
            if ((cc=PositionerMoveAbsolute(motors[mot].hPositioner,motors[mot].axis,tpos,motors[mot].rotCount)) != 0) ccerr(cc);
            isMoving = TRUE;
            calibrStep[mot]++;
            ReleaseSystemMutex(hAttoMutex);
            Sleep(1000);
          }
          else if (calibrStep[mot] == 1)
          { // finished search in one direction and still not valid
            dbglog("%.32s calibration: reverse direction",motors[mot].name);
            calibrDir[mot] *= -1; // switch directions
            calibrStep[mot] = 2;
            stopMoveCmd(mot); // stop moving
            Sleep(1000);
          }
          else
          { // searched both directions and still not valid !
            dbglog("%.32s calibration: searched both directions without finding reference",motors[mot].name);
            motors[mot].status &= ~MTR_UNCLBR; // stop the process
            stopMoveCmd(mot); // stop moving
            calibrStep[mot] = 0;
          }
        }
      }
      else
      { // is valid
        if (calibrStep[mot] < 10)
        {
          stopMoveCmd(mot); // officially stop moving
          Sleep(1000);
          if (WaitForSystemMutex(hAttoMutex,-1) != 0) goto bow;
          dbglog("%.32s calibration: move to reference",motors[mot].name);
          if ((cc=PositionerMoveReference(motors[mot].hPositioner,motors[mot].axis)) != 0) ccerr(cc);
          calibrStep[mot] = 10;
          ReleaseSystemMutex(hAttoMutex);
          Sleep(1000);
        }
        else if (!isMoving)
        { // reference valid, not moving and have moved to reference
          Sleep(1000);
          if (calibrStep[mot] == 10)
          {
            if (WaitForSystemMutex(hAttoMutex,-1) != 0) goto bow;
            dbglog("%.32s calibration: reset position",motors[mot].name);
            if ((cc=PositionerResetPosition(motors[mot].hPositioner,motors[mot].axis)) != 0) ccerr(cc);
            motors[mot].status |= MTR_ISCLBR;
            motors[mot].pos = motors[mot].tgt = 0;
            calibrStep[mot]++;
            ReleaseSystemMutex(hAttoMutex);
            goto bow; // believe it or not you still have to check the valid flag one more time ! :
          }
err:
          dbglog("%.32s finished calibration (status %d)",motors[mot].name,cc);
          if (cc) ReleaseSystemMutex(hAttoMutex);
          motors[mot].status &= ~MTR_UNCLBR; // stop the process
          calibrStep[mot] = 0;
          calibrDir[mot] = 1;
          calibratingMotor = -1;
        }
      }
      goto bow;
    }
    // official moving status bit (sometimes it works!)
    if (motors[mot].stsRb & 0x01) motors[mot].status |= MTR_MOVE; else motors[mot].status &= ~MTR_MOVE;
    // hardcore detection of 'have I reached my target' ...
    if ((motors[mot].status & MTR_MOVE) && fabs(motors[mot].pos - motors[mot].tgt) < MTR_ZERO_TOLERANCE)
    { // officially stop moving
	    if (attoDebugLevel) dbglog("motor %d has reached target %g -> send stop",mot,motors[mot].tgt);
	    moveto(mot,PROC_STS_STOP);
      isMoving = FALSE;
    }
bow:
    Sleep(MTR_UPDATE_INTERVAL);
  }
  return 0;
}
THRHANDLE motorCreateThread(int mot)
{
  DWORD dwThreadId;
  return (THRHANDLE)CreateThread(NULL,0,motorThreadTask,(THREAD_PARAM)mot,0,&dwThreadId);
}

void loadExtLib(char *lib)
{
  char tmp[256], xlib[256];
# if defined(WIN32)
  LPVOID lpMsgBuf;
  char *libExt = ".dll";
  char *libPre = NULL;
# elif defined(UNIX)
  char *libExt = ".so";
  char *libPre = "lib";
# endif
  
  if (lib != NULL && strlen(lib) != 0)
  { /* the external motor lib */
    if (libPre != NULL && strncmp(libPre,xlib,strlen(libPre))) 
    {
      strncpy(tmp,lib,64);
      strcpy(xlib,libPre);
      strncat(xlib,tmp,128);
    }
    if (strcmp(&xlib[strlen(xlib)-strlen(libExt)],libExt)) strcat(xlib,libExt);
    feclog("load motor library %.256s",xlib);
#   if defined(WIN32)
    if (LoadLibrary(xlib) == NULL)
    {
      FormatMessage( 
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,0,NULL);
      feclog("could not load motor library %s : %s",xlib,lpMsgBuf);
    }
#   else
    dlopen(xlib, RTLD_LAZY);
#   endif
  }
}

int serverInitialized = FALSE;
void initialize(void)
{
  PositionerInfo * pi = 0;
  int i, k, cc, n, tocnts=0,nmotorsFromFile=0, rpos=-1,tries=0;
  bool valid=false;
  char errstr[64];
 
  if ((cc=csvReadFile(NULL,CONFIG_FILE,&csvAttoConfigDb,(void **)&motorInfo)) != 0) 
  {
    feclog("cannot read startup configuration parameters from %s : %s",
      CONFIG_FILE,GetSystemErrorString((short)cc,errstr));
    exit(1);
  }
  nmotorsFromFile = csvAttoConfigDb.siz;
 
  if (nmotorsFromFile < 1)
  {
    feclog("no startup configuration parameters from %s given",CONFIG_FILE);
    exit(1);
  }
 
  gNumMotors = nmotorsFromFile;
  if ((motors=(AttoMotor *)calloc(gNumMotors,sizeof(AttoMotor))) == NULL)
  {
    feclog("cannot allocate space for %d motors",gNumMotors);
    exit(1);
  }
  if ((mstNames=(NAME64 *)calloc(gNumMotors,sizeof(NAME64))) == NULL)
  {
    feclog("cannot allocate motor name list ! (aborting ...)");
    exit(1);
  }
  if ((calibrStep=(int *)calloc(gNumMotors,sizeof(int))) == NULL)
  {
    feclog("cannot allocate calibration resources ! (aborting ...)");
    exit(1);
  }
  if ((cc=InitMutex(&hAttoMutex)) != 0)
  {
    feclog("cannot initialize mutex! (aborting ...)");
    exit(1);
  }
  if ((cc=InitMutex(&hCalbrMutex)) != 0)
  {
    feclog("cannot initialize mutex! (aborting ...)");
    exit(1);
  }
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) 
  {
    feclog("cannot get atto mutex! (aborting ...)");
    exit(1);
  }
  for (i=0; i<gNumMotors; i++)
  {
    if (strlen(motorInfo[i].name) == 0) break;
    for (k=0; k<i && strnicmp(motorInfo[i].lib,motorInfo[k].lib,64); k++);
    if (k == i) loadExtLib(motorInfo[i].lib);
    strncpy(mstNames[i].name,motorInfo[i].name,32);
    strncpy(motors[i].name,motorInfo[i].name,32);
    strncpy(motors[i].cfg,motorInfo[i].cfg,64);
    motors[i].mstepsPerCycle *= motors[i].microStepsPerStep; // ??
    for (n=0; n<nmotorsFromFile; n++)
    {
      if (motorInfo[n].motorNr == i+1)
      {
        motors[i].positioner = motorInfo[n].positioner;
        motors[i].axis = motorInfo[n].axis;
        motors[i].hPositioner = getPositionerHandle(motors[i].positioner);
        if (motors[i].hPositioner < 0) continue;
	      if (PositionerSetOutput(motors[i].hPositioner,motors[i].axis,true) == 0)
        {
          motors[i].enabled = TRUE;
        }
        if (strlen(motors[i].cfg) > 0)
        {
          tries = 0;
          while ((cc=PositionerLoad(motors[i].hPositioner,motors[i].axis,motors[i].cfg)) != 0)
          {
            if (++tries > 3) 
            {
              feclog("cannot load %.32s positioner config file %.64s: <%d> %.32s",
                 motors[i].name,motors[i].cfg,cc,GetSystemErrorString((short)ncb2cc(cc),errstr));
              break;
            }
          }
        }
        tries = 0;
        while ((cc=PositionerAmplitudeControl(motors[i].hPositioner,motors[i].axis,2)) != 0)
        {
          if (++tries > 3)
          {
            feclog("cannot set %.32s amplitude control mode: <%d> %.32s",
              motors[i].name,cc,GetSystemErrorString((short)ncb2cc(cc),errstr));
            break;
          }
        }
        tries = 0;
        while ((cc=PositionerGetAmplitude(motors[i].hPositioner,motors[i].axis,&motors[i].amplitude)) != 0)
        {
          if (++tries > 3)
          {
            feclog("cannot get %.32s amplitude : <%d> %.32s",
              motors[i].name,cc,GetSystemErrorString((short)ncb2cc(cc),errstr));
            break;
          }
        }
        tries = 0;
        while ((cc=PositionerGetFrequency(motors[i].hPositioner,motors[i].axis,&motors[i].frequency)) != 0)
        {
          if (++tries > 3)
          {
            feclog("cannot get %.32s frequency : <%d> %.32s",
              motors[i].name,cc,GetSystemErrorString((short)ncb2cc(cc),errstr));
            break;
          }
        }
        tries = 0;
        while ((cc=PositionerGetRotCount(motors[i].hPositioner,motors[i].axis,&motors[i].rotCount)) != 0)
        {
          if (++tries > 3)
          {
            feclog("cannot get %.32s rotation count : <%d> %.32s",
              motors[i].name,cc,GetSystemErrorString((short)ncb2cc(cc),errstr));
            break;
          }
        }
      //	PositionerGetReference(motors[i].hPositioner,motors[i].axis,&rpos,&valid);
      //  motors[i].status |= MTR_UNCLBR;
        strncpy(motors[i].units,motorInfo[n].units,16);
        motors[i].motid = motorInfo[n].motorNr;
        motors[i].maxPos = motorInfo[n].posMax;
        motors[i].minPos = motorInfo[n].posMin;
        motors[i].offPos = motorInfo[n].posOff;
        motors[i].scalePos = motorInfo[n].posScale;
        motors[i].velMean = motorInfo[n].velNormal;
        motors[i].microStepsPerStep = motorInfo[n].mstepsStep;
        motors[i].mstepsPerCycle = motorInfo[n].mstepsCycle;
        motors[i].mstepsPerMM = motorInfo[n].mstepsMM;
        motors[i].cyclesPerMM = motorInfo[n].mstepsMM/motorInfo[n].mstepsCycle;
        motors[i].type = getMotorType(motorInfo[n].type);
        motors[i].rbTolerance = motorInfo[n].rbTolerance;
        motors[i].velLwr = motorInfo[n].velMin;
        motors[i].velUpr = motorInfo[n].velMax;
        motors[i].masterGearRatio = motorInfo[n].gearRatio;
        break;
      }
    }
  }
  ReleaseSystemMutex(hAttoMutex);

  // start the update threads ...
  for (i=0; i<gNumMotors; i++) startUpdateMotor(i);

  //TODO: this double check doesn't work at the moment ...
  n = 0; for (i=0; i<nAttoPositioners; i++) n += attoPositioner[i].deviceCount;
  if (n != gNumMotors)
  {
    feclog("hardware claimed %d motors, config file claimed %d",n,gNumMotors);
  }
  psAxGrps.idLastScanned = -1; /* unknown at server startup */
  serverInitialized = TRUE;
}
int SimulateData(int start,int range)
{
  /* This routine is useful for test servers  */
  return start + rand()%(range);
}
void cbPosPsAx(int id, int cc)
{
  psAxGrps.lnkStsPosPsAx = cc;
}
void cbTgtPsAx(int id, int cc)
{
  psAxGrps.lnkStsTgtPsAx = cc;
}
void cbStsPsAx(int id, int cc)
{
  int i;
  psAxGrps.lnkStsStsPsAx = cc;
  if (cc == 0)
  {
    for (i=0; i< psAxGrps.numPsAxGroups; i++)
    { //1:moving,2:idle,3:Error
      if (psAxGrps.stsBuffer[i] == 0) psAxGrps.status[i] = MTR_IDLE;
      if (psAxGrps.stsBuffer[i] == 1) psAxGrps.status[i] = MTR_MOVE;
      if (psAxGrps.stsBuffer[i] == 2) psAxGrps.status[i] = MTR_IDLE;
      if (psAxGrps.stsBuffer[i] == 3) psAxGrps.status[i] = MTR_ERR;
    }
  }
}
 
void startUpdateMotor(int mot)
{
  if (mot < 0 || mot >= gNumMotors) return;
  motors[mot].hIoThread = motorCreateThread(mot);
}
 
LIB_EXPORT int getUnits(int mot,char *unitsBuffer,int bufferLength)
{
  int len;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  len = MIN(bufferLength,16);
  strncpy(unitsBuffer,motors[mot].units,len);
  return 0;
}
LIB_EXPORT int getPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].pos * motors[mot].scalePos + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getGrpPos(int grpId,double *val)
{
  int i, k, n=0;
  double v = 0, s;
  if (val == NULL) return invalid_parameter;

  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    if (getGroupPositionFcn != NULL) 
    { // a registered function !
      *val = getGroupPositionFcn(&grps[i].grp);
      return 0;
    }
    for (k=0; k<gNumMotors; k++)
    { // find group partners
      if (IsMotorInGroup(k,&grps[i].grp))
      {
        s = GetGroupMotorScale(grpId,k);
        if (s == 0) continue;
        v += motors[i].pos / s;
        n++;
      }
    }
    if (v > 0) v /= n;
    *val = v;
    return 0;
  }
  /*
    switch (grpId)
    {
      case 512://Group motor1 Position
        *val = fabs(motors[0].pos + motors[1].pos);
        break;
      case 513://Group motor2 Position
        *val = fabs(motors[2].pos + motors[3].pos);
        break;
      case 514://Group motor3 Position
        *val = fabs(motors[0].pos - motors[1].pos)/2;
        break;
      case 515://Group motor4 Position
        *val = fabs(motors[2].pos - motors[3].pos)/2;
        break;
      case 516://Group motor5 Position
        *val = fabs(motors[4].pos + motors[5].pos);
        break;
      case 517://Group motor6 Position
        *val = fabs(motors[6].pos + motors[7].pos);
        break;
      case 518://Group motor5 Position
        *val = fabs(motors[4].pos - motors[5].pos)/2;
        break;
      case 519://Group motor6 Position
        *val = fabs(motors[6].pos - motors[7].pos)/2;
        break;
    }
    return 0;
	}
  */
  return invalid_index;
}
LIB_EXPORT int getPosMin(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].minPos * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getPosMax(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].maxPos * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getPosOff(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].offPos;
  return 0;
}
LIB_EXPORT int setPosOff(int mot,double val)
{
  int cc;
  char errstr[64];
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  motors[mot].offPos = (float)val;
  // save and restore:
  motorInfo[mot].posOff = motors[mot].offPos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvAttoConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getPosScale(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = (double)motors[mot].scalePos;
  return 0;
}
LIB_EXPORT int setPosScale(int mot,double val)
{
  int cc;
  char errstr[64];
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  motors[mot].scalePos = (float)val;
  // save and restore:
  motorInfo[mot].posScale = motors[mot].scalePos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvAttoConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getVel(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].velMean;
  return 0;
}
//TODO: trap velocity > upr limit or < lwr limit with software ?
LIB_EXPORT int setVel(int mot,double val)
{
  int cc = not_implemented;
  char errstr[64];
 
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (val < motors[mot].velLwr ||
      val > motors[mot].velUpr) return out_of_range;
  if (cc != 0) return cc;
  // TODO: anything ?
  motors[mot].velMean = (float)(val);
  motorInfo[mot].velNormal = motors[mot].velMean;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvAttoConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getVelLwr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].velLwr;
  return 0;
}
LIB_EXPORT int getVelUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].velUpr;
  return 0;
}
LIB_EXPORT int getAccUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].accUpr;
  return 0;
}
 
// TODO:
LIB_EXPORT int getAcc(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val =  SimulateData(5,2);
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int mstepsPerMM(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].mstepsPerMM;
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int mstepsPerStep(int mot,int *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].microStepsPerStep;
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int stepsPerRadian(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *val = motors[mot].mstepsPerCycle/6.2832;
  return 0;
}
 
int sendMoveCmd(int mot,int useEndSwitch)
{
  int cc, tgt;
# ifdef CRAZY_STUFF
  int i, sts, pos;
# endif
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  tgt = (int) motors[mot].tgt; // steps to mm ?
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
# ifdef CRAZY_STUFF
  Sleep(100);
  for (i=0; i<3; i++)
  {
    PositionerSetTargetPos(motors[mot].hPositioner,motors[mot].axis,tgt,motors[mot].rotCount);ur
    Sleep(100);ur
    cc =PositionerMoveAbsolute(motors[mot].hPositioner,motors[mot].axis,tgt,motors[mot].rotCount);
# endif

    cc = PositionerMoveAbsolute(motors[mot].hPositioner,motors[mot].axis,tgt,0);

# ifdef CRAZY_STUFF
    Sleep(2000);
    cc = PositionerGetStatus(motors[mot].hPositioner,motors[mot].axis,&sts);
    cc = PositionerGetPosition(motors[mot].hPositioner,motors[mot].axis,&pos);
    if ((sts & 0x01) == 0)
    { // atto claims not moving
      if (fabs((double)pos - motors[mot].tgt)!=0) //> MTR_ZERO_TOLERANCE) 
      { // but we're not there yet !
        continue;
      }
    }
    break; // okay to break out 
  }
# endif
  ReleaseSystemMutex(hAttoMutex);
  return ncb2cc(cc);
}
int stopMoveCmd(int mot)
{
  int cc;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
  cc = PositionerStopMoving(motors[mot].hPositioner,motors[mot].axis);
  ReleaseSystemMutex(hAttoMutex);
  return ncb2cc(cc);
}
//TODO: fix
int sendGrpMoveCmd(int gid,int useEndSwitch)
{
// TODO: fix
  return 0;
}
//TODO: fix
int stopGrpMoveCmd(int gid)
{
// TODO: fix
  return 0;
}
//TODO: fix or remove
LIB_EXPORT int scanprg(int mot,int *prg)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (prg == NULL) return invalid_parameter;
  *prg = motors[mot].movEtr;
  return 0;
}
//TODO:  scan in the motor plug ?
#define SCANLENGTH 5000
float scanbuffer[SCANLENGTH*2+4];
int scanlength = SCANLENGTH*2 + 1;
LIB_EXPORT int scansts(int mot,DTYPE *dout)
{
  int i, dlen, cc;
  float *vals;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (dout == NULL) return invalid_parameter;
  if (dout->dArrayLength == 1)
  {
    return PutValuesFromLong(dout,&motors[mot].movCmd,1);
  }
  if (dout->dFormat != CF_FLOAT) return illegal_format;
// TODO: fix
  return 0;

  if (cc) return cc;
  // 1st element contains 1 + 2 x array length
  dlen = (int)scanbuffer[0];
  dlen--;  // subtract the 1st element
  if (dlen < 0) return invalid_data;
  vals = dout->data.fptr;
  if (dlen > (int)dout->dArrayLength) dlen = (int)dout->dArrayLength;
  dout->dArrayLength = dlen;
  for (i=0; i<dlen; i++)
  {
    vals[i] = scanbuffer[1+i];
  }
  return 0;
}
LIB_EXPORT int scanto(int mot,int cmd,DTYPE *din)
{
  double etr;
  int cc,sts;
  double dvals[32];
  static int dlen = 0;
// TODO: fix
  return 0;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
//  sprintf(dev,"%.32s.StepScan",motors[mot].name);
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      dlen = MIN(din->dArrayLength,32);
      if ((cc=GetValuesAsDouble(din,dvals,32)) != 0) return cc;
//     if (sendDoubleArrayToHardware(dev,dvals,dlen) == 0)
      if (TRUE)
      {
        //TODO: fix this
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      break;
    case PROC_STS_RESUME:
 //     if (sendDoubleArrayToHardware(dev,dvals,dlen) == 0)
      {
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
  }
  return 0;
}

LIB_EXPORT int movests(int mot,int *sts)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (sts == NULL) return invalid_parameter;
  *sts = motors[mot].movCmd;
  return 0;
}
LIB_EXPORT int moveprg(int mot,int *prg)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (prg == NULL) return invalid_parameter;
  *prg = motors[mot].movEtr;
  return 0;
}
int sendStepCmd(int mot,int useEndSwitch)
{
  int steps = 0;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if ((motors[mot].status & (MTR_CLBR|MTR_MOVE)) != 0) return not_allowed;

  motors[mot].moving = FALSE;
  motors[mot].movpos = FALSE;
  motors[mot].movneg = FALSE;
//  steps = (int)((motors[mot].tgt - motors[mot].pos)/(double)motors[mot].nmPerStep);
  //steps = motors[mot].nsteps;
  if (steps == 0) steps = 1; // no input -> just step forward

  motors[mot].status |= MTR_STEP;
  motors[mot].stepping = TRUE;
  //motors[mot].nsteps = steps;
  return 0;
}
LIB_EXPORT int stepto(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      //motors[mot].moveStart = 0;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendStepCmd(mot,TRUE) == 0)
      {
        //motors[mot].moveStart = makeDataTimeStamp();
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      //motors[mot].moveStart = makeDataTimeStamp();
      break;
    case PROC_STS_RESUME:
      if (sendStepCmd(mot,TRUE) == 0)
      {
        //motors[mot].moveStart = makeDataTimeStamp();
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      //motors[mot].moveStart = 0;
      break;
  }
  return 0;
}
LIB_EXPORT int moveto(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendMoveCmd(mot,TRUE) == 0)
      {
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        //etr /= (double)(motors[mot].velRps * motors[mot].mstepsPerCycle);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      break;
    case PROC_STS_RESUME:
      if (sendMoveCmd(mot,TRUE) == 0)
      {
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
  }
  return 0;
}
//TODO: test ?
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  switch (cmd)
  {
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendMoveCmd(mot,FALSE) == 0)
      {
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    default:
      return moveto(mot,cmd);
  }
  return 0;
}
int getGrpTblIndex(MotorGroup *grp) 
{ 
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (memcmp(&grps[i].grp,grp,sizeof(MotorGroup))) continue;
    return i;
  }
  return -1; 
}
int addToGrpTbl(MotorGroup *grp) 
{ 
  int i = getGrpTblIndex(grp);
  if (i >= 0) return i;
  if (ngrps < GRPTBL_SIZE)
  {
    i = ngrps;
    grps[i].gearRatio = 0;
    memcpy(&grps[i].grp,grp,sizeof(MotorGroup));
    ngrps++;
    return i;
  }
  return -1; 
}
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp)
{
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    memcpy(grp,&grps[i].grp,sizeof(MotorGroup));
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int setGrpMembers(int grpId,MotorGroup *grp)
{
  int g;
  if (grpId < MST_DYNAGRP_OFFST)
  {
    if ((g=getPsAxGroupIndex(grpId)) < 0)
    {
      if (psAxGrps.numPsAxGroups >= PSAX_BUFFER_SIZE) return resources_exhausted;
      g = psAxGrps.numPsAxGroups;
      psAxGrps.numPsAxGroups++;
    }
    psAxGrps.grpIdXref[g] = grpId;
  }
  else
  {
    if ((g=addToGrpTbl(grp)) < 0) return invalid_index;
    grps[g].id = grpId;
  }
//  memcpy(&grps[g].grp,grp,sizeof(MotorGroup));
  return 0;
}
LIB_EXPORT int setGrpGearRatio(int grpId,double gearRatio)
{
  int i;
  if (gearRatio == 0) return out_of_range;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    grps[i].gearRatio = (float)gearRatio;
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int getGrpGearRatio(int grpId,double *gearRatio)
{
  int i;
  if (gearRatio == NULL) return invalid_parameter;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    *gearRatio = (double)grps[i].gearRatio;
    return 0;
  }
  return invalid_index;
}
/*
  e.g. 
	if (i == 2)  (tp = (tgt/2) + (motors[2].pos-motors[3].pos)/2);
	if (i == 3)  (tp = (tgt/2) - (motors[2].pos-motors[3].pos)/2);

  scale. weight refer to target motor

  scale = 0.5, weight = 0.5 ...

	512://Group motor1 Position GapHor3
  scale = 0.5, weight = 0.5
  if (motors[i].motid==1)  (tp = (tgt/2) + (motors[0].pos-motors[1].pos)/2);
  if (motors[i].motid==2)  (tp = (tgt/2) - (motors[0].pos-motors[1].pos)/2);
  513://Group motor2 Position GapVer3
	if (motors[i].motid==3)  (tp = (tgt/2) + (motors[2].pos-motors[3].pos)/2);
	if (motors[i].motid==4)  (tp = (tgt/2) - (motors[2].pos-motors[3].pos)/2);
  scale = 0.5, weight = 0.5
  514://Group motor3 Position OffHor3
  if (motors[i].motid==1)  (tp = tgt + (motors[0].pos-motors[1].pos)/4);
  if (motors[i].motid==2)  (tp = tgt - (motors[0].pos-motors[1].pos)/4);
  scale = 1.0, weight = 0.25
  515://Group motor4 Position OffVer3
  if (motors[i].motid==3)  (tp = tgt + (motors[2].pos-motors[3].pos)/4);
  if (motors[i].motid==4)  (tp = tgt - (motors[2].pos-motors[3].pos)/4);
  scale = 1.0, weight = 0.25
	516://Group motor5 Position GapHor4
  if (motors[i].motid==5)  (tp = (tgt/2) + (motors[4].pos-motors[5].pos)/2);
  if (motors[i].motid==6)  (tp = (tgt/2) - (motors[4].pos-motors[5].pos)/2);
  scale = 0.5, weight = 0.5
	517://Group motor6 Position GapVer4
  if (motors[i].motid==7)  (tp = (tgt/2) + (motors[6].pos-motors[7].pos)/2);
  if (motors[i].motid==8)  (tp = (tgt/2) - (motors[6].pos-motors[7].pos)/2);
  scale = 0.5, weight = 0.5
	518://Group motor5 Position OffHor4
  if (motors[i].motid==5)  (tp = tgt + (motors[4].pos-motors[5].pos)/4);
  if (motors[i].motid==6)  (tp = tgt - (motors[4].pos-motors[5].pos)/4);
  scale = 1.0, weight = 0.25
	519://Group motor6 Position OffVer4
  if (motors[i].motid==7)  (tp = tgt + (motors[6].pos-motors[7].pos)/4);
  if (motors[i].motid==8)  (tp = tgt - (motors[6].pos-motors[7].pos)/4);
  scale = 1.0, weight = 0.25

*/
float getGroupMemberTargetPosition(MotorGroup *grp, int mot, float target)
{ // sum over : target * scale + ((motor->pos) - (partner->pos)) * weight
  int i;
  float scale, weight, tp;
  // a registered external functin ?
  if (getGroupMemberTargetPositionFcn != NULL) return getGroupMemberTargetPositionFcn(grp,mot,target);
  // then do it this way ...
  scale = GetGroupMotorScale(grp->grpid,mot);
  weight = GetGroupMotorWeight(grp->grpid,mot);
  tp = target * scale;
  for (i=0; i<gNumMotors; i++)
  { // find group partners
    if (i == mot) continue;
    if (IsMotorInGroup(i,grp))
    { // this one is a group member !

      // motors[i].pos or motors[i].tgt ???

      tp += (float)(motors[mot].pos - motors[i].pos) * weight;
    }
  }
  return tp;
}

extern "C"
{
LIB_EXPORT void setGetGroupMemberTargetPositionFunction(GETGMTPFCNP fcn)
{
  getGroupMemberTargetPositionFcn = fcn;
}
LIB_EXPORT int hasExternGetGroupMemberTargetPositionFunction(void)
{
  return getGroupMemberTargetPositionFcn == NULL ? FALSE : TRUE;
}
LIB_EXPORT void setGetGroupPositionFunction(GETGPFCNP fcn)
{
  getGroupPositionFcn = fcn;
}
LIB_EXPORT int hasGroupPositionPositionFunction(void)
{
  return getGroupPositionFcn == NULL ? FALSE : TRUE;
}
}
//TODO: fix
LIB_EXPORT int grpmoveto(int grpId,int cmd)
{
  int cc=0, i, gidx, g=-1;
  MotorGroup grp;
  double etr, gearRatio=0;
  int tgt=0, tp=0, mstart, mnext, hPos, axes;

  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  if (grp.grphead > 0) 
  { // group master defined from above
    g = grp.grphead - 1;
  }
  else
  { // not defined: take the first in the list
    for (i=0; i<gNumMotors; i++) if (IsMotorInGroup(i,&grp)) break;
    if (i == gNumMotors) return configuration_error;
    g = i;
  }
  if (cmd == PROC_STS_START)
  { // first things first
    for (i=0; i<gNumMotors; i++)
    { 
      if (IsMotorInGroup(i,&grp))
      { // don't allow if anybody is being calibrated or is moving !
        if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
        if ((motors[i].status & MTR_MOVE) != 0) return operation_busy;
        if (g == i) 
        { // fix the coupling parameters 
          getGrpGearRatio(grpId,&gearRatio);
          if (gearRatio == 0) gearRatio = (double)motors[i].masterGearRatio;
          if (tgt == 0) tgt = (int)motors[i].tgt;
        }
      }
    }
    for (i=0; i<gNumMotors; i++)
    { // couple all motors in group to same axis
      if (IsMotorInGroup(i,&grp))
      { // this one is a group member !      
        motors[i].grpId = grpId; // make a note of the group id
        gidx = getGrpTblIndex(&grp);
        tp = gidx >= 0 ? (int)getGroupMemberTargetPosition(&grp,i,(float)tgt) : tgt;
        if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
        cc = PositionerSetTargetPos(motors[i].hPositioner,motors[i].axis,tp,motors[i].rotCount);
        ReleaseSystemMutex(hAttoMutex);
        if (cc == 0) motors[i].tgt = tp; else dbglog("set target error %d",cc);
      }
    }
  }
  for (i=0; i<gNumMotors; i++)
  { // send move commands only to the master
    if (IsMotorInGroup(i,&grp))
    {       
      int lastHpos = -1;
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          motors[i].grpId = -1;
          break;
        case PROC_STS_RESUME:
        case PROC_STS_START:
          mstart = 0; mnext = -1; hPos = -1; axes = 0;
GentlemenStartYourMotors:
          for (i=mstart; i<gNumMotors; i++)
          { // find all axes to a positioner
            if (IsMotorInGroup(i,&grp))
            { // this one is a group member !
              if (hPos < 0 && motors[i].hPositioner != lastHpos) hPos = motors[i].hPositioner; 
              if (hPos == motors[i].hPositioner)
              {
                axes |= 0x01 << motors[i].axis;
              }
              else 
              { // next round
                mnext = i;
              }
            }
#           ifndef USE_ABSOLUTESYNC
            sendMoveCmd(i,TRUE);
#           endif
            etr = fabs(motors[i].tgt - motors[i].pos);
            etr /= (double)(motors[i].velMean);
            //etr /= (double)(motors[i].velRps * motors[i].mstepsPerCycle);
            motors[i].movEtr = (int)etr;
            motors[i].movPaused = FALSE;
            motors[i].status |= MTR_MOVE;
          }
#         ifdef USE_ABSOLUTESYNC
          if (hPos >= 0)
          {
            if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
            PositionerMoveAbsoluteSync(hPos,axes);
           // Sleep(500);uwe
            ReleaseSystemMutex(hAttoMutex);
            lastHpos = hPos;
            hPos = -1; axes = 0;
            mstart = mnext < 0 ? gNumMotors : mnext;
            goto GentlemenStartYourMotors;
          }
#         endif
          break;
        case PROC_STS_PAUSE:
          if (i == g) stopMoveCmd(g);
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          for (i=0; i<gNumMotors; i++)
          { // find all axes to a positioner
            if (IsMotorInGroup(i,&grp))
            { // this one is a group member !
              if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
              PositionerStopMoving(motors[i].hPositioner,motors[i].axis);
              ReleaseSystemMutex(hAttoMutex);
            }
            motors[i].movEtr = 0;
            motors[i].movPaused = FALSE;
            motors[i].status = MTR_IDLE;
          }
          break;
      }
    }
  }
  if (cmd == PROC_STS_STOP)
  { // finished
    // TODO: fix this
    for (i=0; i<gNumMotors; i++)
    { // decouple all coupled motors from axis
      if (IsMotorInGroup(i,&grp))
      {       
        if (i == g) continue;
      }
    }
  }
  return cc;
}
LIB_EXPORT int incgrpmoveto(int grpId,int cmd)
{
  int cc=0, i, gidx, g=-1;
  MotorGroup grp;
  double etr, gearRatio=0;
  int tgt=0, tp, mstart, mnext, hPos, axes;

  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  if (grp.grphead > 0) 
  { // group master defined from above
    g = grp.grphead - 1;
  }
  else
  { // not defined: take the first in the list
    for (i=0; i<gNumMotors; i++) if (IsMotorInGroup(i,&grp)) break;
    if (i == gNumMotors) return configuration_error;
    g = i;
  }
  if (cmd == PROC_STS_START)
  { // first things first
    for (i=0; i<gNumMotors; i++)
    { 
      if (IsMotorInGroup(i,&grp))
      { // don't allow if anybody is being calibrated or is moving !
        if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
        if ((motors[i].status & MTR_MOVE) != 0) return operation_busy;
        if (g == i) 
        { // fix the coupling parameters 
          getGrpGearRatio(grpId,&gearRatio);
          if (gearRatio == 0) gearRatio = (double)motors[i].masterGearRatio;
          if (tgt == 0) 
          { // find the increment again (until we do this right)
            tgt = (int)(motors[i].tgt - motors[i].pos);
          }
        }
      }
    }
    for (i=0; i<gNumMotors; i++)
    { // couple all motors in group to same axis
      if (IsMotorInGroup(i,&grp))
      { // this one is a group member !      
        motors[i].grpId = grpId; // make a note of the group id
        gidx = getGrpTblIndex(&grp);
        tp = tgt;
        if (gidx >= 0)
        {
          gearRatio = GetGroupMotorGearRatio(grpId,i);
	        tp = (int)motors[i].pos + (int)(tgt * gearRatio);
	        if (attoDebugLevel) dbglog("motor %d: gear ratio %g, pos %g, final target pos %d",i,gearRatio,motors[i].pos,tp);
        }
        if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
        cc = PositionerSetTargetPos(motors[i].hPositioner,motors[i].axis,tp,motors[i].rotCount);
        ReleaseSystemMutex(hAttoMutex);
        if (cc == 0) motors[i].tgt = tp; else dbglog("set target error %d",cc);
      }
    }
  }
  for (i=0; i<gNumMotors; i++)
  { // send move commands only to the master
    if (IsMotorInGroup(i,&grp))
    {       
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          motors[i].grpId = -1;
          break;
        case PROC_STS_RESUME:
        case PROC_STS_START:
          mstart = 0; mnext = -1; hPos = -1; axes = 0;
GentlemenStartYourMotors:
          for (i=mstart; i<gNumMotors; i++)
          { // find all axes to a positioner
            if (IsMotorInGroup(i,&grp))
            { // this one is a group member !
              if (hPos < 0) hPos = motors[i].hPositioner; 
              if (hPos == motors[i].hPositioner)
              {
                axes |= 0x01 << motors[i].axis;
              }
              else 
              { // next round
                mnext = i;
              }
            }
#           ifndef USE_ABSOLUTESYNC
            sendMoveCmd(i,TRUE);
#           endif
            etr = fabs(motors[i].tgt - motors[i].pos);
            etr /= (double)(motors[i].velMean);
            //etr /= (double)(motors[i].velRps * motors[i].mstepsPerCycle);
            motors[i].movEtr = (int)etr;
            motors[i].movPaused = FALSE;
            motors[i].status |= MTR_MOVE;
          }
#         ifdef USE_ABSOLUTESYNC
          if (hPos >= 0)
          {
            if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
            PositionerMoveAbsoluteSync(hPos,axes);
            ReleaseSystemMutex(hAttoMutex);
            hPos = -1; axes = 0;
            mstart = mnext < 0 ? gNumMotors : mnext;
            goto GentlemenStartYourMotors;
          }
#         endif
          break;
        case PROC_STS_PAUSE:
          if (i == g) stopMoveCmd(g);
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          for (i=0; i<gNumMotors; i++)
          { // find all axes to a positioner
            if (IsMotorInGroup(i,&grp))
            { // this one is a group member !
              if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
              PositionerStopMoving(motors[i].hPositioner,motors[i].axis);
              ReleaseSystemMutex(hAttoMutex);
            }
            motors[i].movEtr = 0;
            motors[i].movPaused = FALSE;
            motors[i].status = MTR_IDLE;
          }
          break;
      }
    }
  }
  if (cmd == PROC_STS_STOP)
  { // finished
    // TODO: fix this
    for (i=0; i<gNumMotors; i++)
    { // decouple all coupled motors from axis
      if (IsMotorInGroup(i,&grp))
      {       
        if (i == g) continue;
      }
    }
  }
  return cc;
}
//TODO : remove or fix
LIB_EXPORT int grpscansts(int grpId,DTYPE *dout)
{
  int i, mot, dlen, cc;
  float *vals;
  MotorGroup grp;
  //TODO GROUP ???
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors && !IsMotorInGroup(i,&grp); i++);
  if (i == gNumMotors) return illegal_device;
  mot = i;
  if (dout == NULL) return invalid_parameter;
  if (dout->dArrayLength == 1)
  {
    return PutValuesFromLong(dout,&motors[mot].movCmd,1);
  }
  if (dout->dFormat != CF_FLOAT) return illegal_format;
//  sprintf(dev,"%.32s.ScanArray",motors[mot].name);
//getScanData:
//  cc = getFloatArrayFromHardware(dev,scanbuffer,scanlength);
  cc = not_implemented;
  if (cc) return cc;
  dlen = (int)scanbuffer[0]; dlen--;
  vals = dout->data.fptr;
  if (dlen > (int)dout->dArrayLength) dlen = (int)dout->dArrayLength;
  dout->dArrayLength = dlen;
  dlen /= 2;
  for (i=0; i<dlen; i++)
  {
    vals[i] = scanbuffer[1+i];
    vals[i+dlen] = scanbuffer[1+dlen+i];
  }
  return 0;
}
//TODO : remove or fix
LIB_EXPORT int grpscanto(int grpId,int cmd,DTYPE *din)
{
  int cc, i, g=-1;
  MotorGroup grp;
  double etr, gearRatio;
  float fvals[4];
  static double dvals[32];
  static int dlen = 0;

  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  if (grp.grphead > 0) 
  { // group master defined from above
    g = grp.grphead - 1;
  }
  else
  { // not defined: take the first in the list
    for (i=0; i<gNumMotors; i++) if (IsMotorInGroup(i,&grp)) break;
    if (i == gNumMotors) return configuration_error;
    g = i;
  }
  if (cmd == PROC_STS_START)
  { // first things first
    for (i=0; i<gNumMotors; i++)
    { 
      if (IsMotorInGroup(i,&grp))
      { // don't allow if anybody is being calibrated or is moving !
        if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
        if ((motors[i].status & MTR_MOVE) != 0) return operation_busy;
        if (g == i) 
        { // fix the coupling parameters 
          getGrpGearRatio(grpId,&gearRatio);
          if (gearRatio == 0) gearRatio = (double)motors[i].masterGearRatio;
          fvals[0] = 1;
          fvals[1] = (float)(g + 1);
          fvals[2] = (float)gearRatio;
        }
      }
    }
    for (i=0; i<gNumMotors; i++)
    { // couple all motors in group to same axis
      if (IsMotorInGroup(i,&grp))
      { // this one is a group member !      
        motors[i].grpId = grpId; // make a note of the group id
        if (g == i) continue; // don't couple the master
      }
    }
  }  
  for (i=0; i<gNumMotors; i++)
  { // send move commands only to the master
    if (IsMotorInGroup(i,&grp))
    {       
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
        case PROC_STS_START:
          if (i == g) scanto(g,cmd,din);
          etr = fabs(motors[i].tgt - motors[i].pos);
          etr /= (double)(motors[i].velMean);
          motors[i].movEtr = (int)etr;
          motors[i].movPaused = FALSE;
          motors[i].status |= MTR_MOVE;
          break;
        case PROC_STS_PAUSE:
          if (i == g) stopMoveCmd(g);
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_RESUME:
          if (i == g) scanto(g,cmd,din);
          motors[i].movPaused = FALSE;
          motors[i].status &= ~MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          if (i == g) stopMoveCmd(g);
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
      }
    }
  }
  if (cmd == PROC_STS_STOP)
  { // finished
    for (i=0; i<gNumMotors; i++)
    { // decouple all coupled motors from axis
      if (IsMotorInGroup(i,&grp))
      {       
        if (i == g) continue;
      }
    }
  }
  return 0;
}
LIB_EXPORT int setPos(int mot,double val)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (val < motors[mot].minPos || val > motors[mot].maxPos) 
  {
    if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
  }
  //TODO: fill this in !
  return 0;
}
LIB_EXPORT int setTgtPos(int mot,double val)
{
  int cc, tgt;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (val < motors[mot].minPos || val > motors[mot].maxPos) 
  {
    if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
  }
  tgt = (int)val; 
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
  cc = PositionerSetTargetPos(motors[mot].hPositioner,motors[mot].axis,tgt,motors[mot].rotCount);
  ReleaseSystemMutex(hAttoMutex);
  if (cc == 0) motors[mot].tgt = val;  
  return ncb2cc(cc);
}
LIB_EXPORT int getTgtPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
 
  *val = motors[mot].tgt * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int setGrpTgtPos(int grpId,double val)
{
  int cc, i, g=-1;
  MotorGroup grp;

  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors; i++)
  { // don't allow if anybody is being calibrated !
    if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
  }
  for (i=0; i<gNumMotors && g < 0; i++) if (IsMotorInGroup(i,&grp)) g = i;
  return setTgtPos(g,val);
}
LIB_EXPORT int getGrpTgtPos(int grpId,double *val)
{
  int cc, i, g=-1;
  MotorGroup grp;

  if (val == NULL) return invalid_parameter;
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors && g < 0; i++) if (IsMotorInGroup(i,&grp)) g = i;
  if (g < 0 || g > gNumMotors) return illegal_device;
  return getTgtPos(g,val);
}
//TODO GROUP ???
LIB_EXPORT int calibr(int mot,int *sts)
{
  int cc = 0;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if ((motors[mot].status & 0xff) != MTR_IDLE) return not_allowed;
//  cc = PositionerResetPosition(motors[mot].hPositioner,motors[mot].axis);
  calibrStep[mot] = 0;
  motors[mot].status |= MTR_UNCLBR;
  return 0;
}
//TODO: remove or fix
LIB_EXPORT int calibrToOffset(int mot,double val)
{
  int cc = not_implemented;

  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (motors[mot].type != MTR_TYPE_LINEAR) return not_allowed;
  motors[mot].status |= MTR_ISCLBR;
  return cc;
}
LIB_EXPORT int getonline(int mot,int *sts)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device; 
  if (sts == NULL) return argument_list_error;
  *sts = motors[mot].enabled;
  return 0;
}
LIB_EXPORT int setonline(int mot,int sts)
{
  int cc = 0;
  bool val = sts ? true : false;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
	if ((cc=PositionerSetOutput(motors[mot].hPositioner,motors[mot].axis,val)) == 0)
  {
    motors[mot].enabled = sts;
    cc = ncb2cc(cc);
  }
  ReleaseSystemMutex(hAttoMutex);
  return 0;
}

LIB_EXPORT int reset(int mot,int *sts)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device; 
  motors[mot].status = MTR_UNCLBR;
  calibr(mot,sts);
  return 0;
}
//TODO: necessary ?
int chkEndStation(int mot,int mask)
{
//char dev[128];
  int reg = 0;
//sprintf(dev,"%s.regExSta",motors[mot].reg);
//getIntegerFromHardware(dev,&reg);
  return (reg & mask) ? TRUE : FALSE;
}
LIB_EXPORT int getSts(int mot,int *sts)
{ // is called from the background task @ 10 Hz
  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  *sts = motors[mot].status;
  return 0;
}
LIB_EXPORT int getGrpSts(int grpId,int *sts)
{
  int i, m;
  if (sts == NULL) return invalid_parameter;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    m = grps[i].grp.grphead - 1;
    *sts = motors[m].status;
    return 0;
  }
  return invalid_index;
}
int makeMotStsBits(int mot)
{
  int stsBits=0;
  if (mot < 0 || mot >= gNumMotors) return 0;
  if (motors[mot].idle) stsBits |= STS_IDLE;
  if (motors[mot].enabled) stsBits |= STS_ENABLED;
  if (motors[mot].moving) stsBits |= STS_MOVING;
  if (motors[mot].movpos) stsBits |= STS_MOVPOS;
  if (motors[mot].movneg) stsBits |= STS_MOVNEG;
  if (motors[mot].endpos) stsBits |= STS_ENDPOS;
  if (motors[mot].endneg) stsBits |= STS_ENDNEG;
  return stsBits;
}
LIB_EXPORT int getStsInfo(int mot,MotorStatus *motsts)
{
  if (motsts == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  strncpy(motsts->motName,motors[mot].name,64);
  motsts->stsBits = makeMotStsBits(mot);
  motsts->stsAxis = motors[mot].axisState;
  motsts->stsCoupled = motors[mot].coupledState;
  motsts->stsClbr = motors[mot].clbrSts;
  motsts->stateClbr = motors[mot].clbrState;
  motsts->idxCntlLoop = motors[mot].actCntrlLoopIndex;
  motsts->pos = (float)motors[mot].posRb;
  motsts->tgt = (float)motors[mot].tgtRb;
  motsts->lag = motors[mot].lag;
  motsts->vel = motors[mot].velRb;
  motsts->load = motors[mot].pcntMaxCur;
  motsts->ovrRide = motors[mot].overrideSts;
  return 0;
}

LIB_EXPORT int getLastHdwErr(int mot,NAME64 *errstr,int *errcode)
{
  char str[128];
  int cc = 0, sts;
  if (errstr == NULL && errcode == NULL) return invalid_parameter;
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
  cc = PositionerGetStatus(motors[mot].hPositioner,motors[mot].axis,&sts);
  ReleaseSystemMutex(hAttoMutex);

  switch (cc)
  {
    case NCB_Ok:
      if (sts & 0x01) strcpy(str,"axis moving");
      if (sts & 0x02) strcpy(str,"at limit");
      if (sts & 0x04) strcpy(str,"sensor error");
      if (sts & 0x08) strcpy(str,"sensor not connected");
      break;
    case NCB_DriverError: strcpy(str,"driver error"); // - Treiberproblem, nicht installiert?
      break;
    case NCB_NotConnected: strcpy(str,"not connected: check cable, power"); // - Kein Kontakt zum Gert, USB-Kabel? DC-Power?
      break;
    default:
    case NCB_Error: strcpy(str,"unknown error"); //        - Unklarer Fehler
      break;
  }
  cc = ncb2cc(cc);

  motors[mot].status = sts;

  if (errstr != NULL) 
  { 
    strncpy(errstr->name,str,64);
  }
  if (errcode != NULL)
  { 
    *errcode = cc;
  }
  return 0;
}

XMA xprps[] = { 
  {"Amplitude", "motor DC voltage amplitude", "V", 60000, 0, CA_READ|CA_WRITE},
  {"Frequency", "rotational frequency of motor", "Hz", 1000, 0, CA_READ|CA_WRITE},
  {"RotationCount", "rotation count of motor", "Counts", 256, 0, CA_READ},
  {"isValid", "motor is valid flag", "none", 1, 0, CA_READ},
  {"attoStatus", "motor status readback", "bits", 1, 0, CA_READ},
};
int nxprps = sizeof(xprps)/sizeof(XMA);
#define IDX_AMPL   0
#define IDX_FREQ   1
#define IDX_ROTCNT 2
#define IDX_ISVALID 3
#define IDX_ATTOSTS 4

int getExtendedProperty(char *prpName)
{
  int i;
  for (i=0; i<nxprps; i++) if (!strnicmp(prpName,xprps[i].name,64)) return i;
  return -1;
}
int getMotorFromName(char *name)
{
  int i;
  for (i=0; i<gNumMotors; i++) if (!strnicmp(mstNames[i].name,name,32)) return i;
  return -1;
}
int attoeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int i, cc=0, ival, rpos;
  int prp=getExtendedProperty(devProperty), mot = getMotorFromName(devName);
  int mca[32];
  bool valid;

  if (!serverInitialized) return not_initialized;
  if (prp < 0) return illegal_property;
  if (mot < 0) return illegal_device;
  switch (prp)
  {
    case IDX_AMPL:
      if (access & CA_WRITE)
      {
        if (din->dArrayLength != 1) return dimension_error;
        if ((cc=GetValuesAsLong(din,&ival,1)) != 0) return cc;
        if (ival < xprps[prp].min || ival > xprps[prp].max) return out_of_range;
        if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
        cc = ncb2cc(PositionerAmplitude(motors[mot].hPositioner,motors[mot].axis,ival));
        ReleaseSystemMutex(hAttoMutex);
        if (cc != 0) return cc;
        motors[mot].amplitude = ival;
      }
      if (dout->dArrayLength > 0)
      {
        for (i=0; i<32 && i<gNumMotors; i++) mca[i] = motors[i].amplitude;
        cc = PutValuesFromLongEx(dout,mca,gNumMotors,mot);
      }
      return cc;
    case IDX_FREQ:
      if (access & CA_WRITE)
      {
        if (din->dArrayLength != 1) return dimension_error;
        if ((cc=GetValuesAsLong(din,&ival,1)) != 0) return cc;
        if (ival < xprps[prp].min || ival > xprps[prp].max) return out_of_range;
        if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
        cc = ncb2cc(PositionerFrequency(motors[mot].hPositioner,motors[mot].axis,ival));
        ReleaseSystemMutex(hAttoMutex);
        if (cc != 0) return cc;
        motors[mot].frequency = ival;
      }
      if (dout->dArrayLength > 0)
      {
        for (i=0; i<32 && i<gNumMotors; i++) mca[i] = motors[i].frequency;
        cc = PutValuesFromLongEx(dout,mca,gNumMotors,mot);
      }
      return cc;
    case IDX_ROTCNT:
      if (access & CA_WRITE) return illegal_read_write;
      if (dout->dArrayLength > 0)
      {
        for (i=0; i<32 && i<gNumMotors; i++) mca[i] = motors[i].rotCount;
        cc = PutValuesFromLongEx(dout,mca,gNumMotors,mot);
      }
      return cc;
	case IDX_ISVALID:
      if (access & CA_WRITE) return illegal_read_write;
      if (dout->dArrayLength > 0)
      {
        if (WaitForSystemMutex(hAttoMutex,-1) != 0) return semaphore_busy;
        cc = PositionerGetReference(motors[mot].hPositioner,motors[mot].axis, &rpos,&valid);
        if (cc != 0) return ncb2cc(cc);
        motors[mot].valid = valid ? TRUE : FALSE;
        ReleaseSystemMutex(hAttoMutex);
        cc = PutValuesFromLongEx(dout,&motors[mot].valid,1,0);
      }
      return cc;
	case IDX_ATTOSTS:
      if (access & CA_WRITE) return illegal_read_write;
      if (dout->dArrayLength > 0)
      {
        cc = PutValuesFromLong(dout,&motors[mot].stsRb,1);
      }
      return cc;

    default:
      break;
  }
  return not_implemented;
}
int startup(void)
{
  initialize();
  SetMotorList(mstNames,gNumMotors); // done
  SetXMAList(xprps,nxprps,attoeqm);
  SetStatusList(stsLst,sizStsLst);   // done 
  SetGetMotorStatusDispatch(getSts); // done
  SetGetGroupMotorStatusDispatch(getGrpSts);
  SetGetMotorVelocityDispatch(getVel); // done
  SetSetMotorVelocityDispatch(setVel); // done
  SetGetMotorAccelerationDispatch(getAcc);
  SetGetMotorPositionDispatch(getPos); // done
  SetGetGroupMotorPositionDispatch(getGrpPos);
  SetSetMotorPositionDispatch(setPos);
  SetGetMotorStepsPerMUDispatch(mstepsPerMM); // done
  SetGetMotorMicroStepsPerStepDispatch(mstepsPerStep); // done
  SetGetMotorStepsPerRadianDispatch(stepsPerRadian); // done
  SetGetMoveToStatusDispatch(movests); // done
  SetGetMoveToProgressDispatch(moveprg); // done
  SetGetGroupMoveStatusDispatch(movests);
  SetGetIncrementMoveStatusDispatch(movests);
  SetGetIncrementGroupMoveStatusDispatch(movests);
  SetSetMoveToStatusDispatch(moveto); // done
  SetSetFreeMoveToStatusDispatch(movetoNoEndSwitch); // done
  SetSetGroupMoveStatusDispatch(grpmoveto); // done ?
  SetSetIncrementGroupMoveStatusDispatch(incgrpmoveto);
  SetSetIncrementMoveStatusDispatch(moveto); // done (same as MoveTo)
  //SetSetStepToStatusDispatch(stepto);
  SetSetMotorTargetPositionDispatch(setTgtPos); // done
  SetGetMotorTargetPositionDispatch(getTgtPos); // done
 // SetSetGroupMotorTargetPositionDispatch(setGrpTgtPos);
 // SetGetGroupMotorTargetPositionDispatch(getGrpTgtPos);
  SetGetGroupMembersDispatch(getGrpMembers); // done ?
  SetSetGroupMembersDispatch(setGrpMembers); // done ?
  SetCalibrateMotorDispatch(calibr);  // done
  SetResetMotorDispatch(reset); // done
  SetGetMotorVelocityLowerLimitDispatch(getVelLwr);
  SetGetMotorVelocityUpperLimitDispatch(getVelUpr);
//  SetGetMotorAccelerationUpperLimitDispatch(getAccUpr);
  SetGetMotorPositionMinimumDispatch(getPosMin); // done
  SetGetMotorPositionMaximumDispatch(getPosMax); // done
  SetGetMotorPositionOffsetDispatch(getPosOff); // done
  SetSetMotorPositionOffsetDispatch(setPosOff); // done
  SetGetMotorPositionScaleDispatch(getPosScale); // done
  SetSetMotorPositionScaleDispatch(setPosScale); // done
  SetCalibrateToOffsetDispatch(calibrToOffset);
  SetGetMotorUnitsDispatch(getUnits);
//  SetGetScanMoveToStatusDispatch(scansts);
//  SetSetScanMoveToStatusDispatch(scanto);
//  SetGetScanGroupMoveToStatusDispatch(grpscansts);
//  SetSetScanGroupMoveToStatusDispatch(grpscanto);
//  SetGetScanMoveToProgressDispatch(scanprg);
  SetGetMotorOnlineDispatch(getonline);
  SetSetMotorOnlineDispatch(setonline);
  SetGetMotorStatusBlockDispatch(getStsInfo);
  SetSetGroupGearRatioDispatch(setGrpGearRatio); // done
  SetGetGroupGearRatioDispatch(getGrpGearRatio); // done
  SetGetLastHardwareErrorDispatch(getLastHdwErr);
  // allow access to a debug flag
  RegisterUserCommand("attodebug",NULL,(int *)&attoDebugLevel,NULL,CA_READ|CA_WRITE);

  return 0;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      SetInitMotorHardwareDispatch(startup);
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      attoShutdown();
      CloseSystemMutex(hAttoMutex);
      break;
  }
  return (TRUE);
}


