#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "mstlib.h"
 
#if defined(MSTBEK_EXPORTS) 
# define LIB_EXPORT __declspec(dllexport)
# else
# define LIB_EXPORT __declspec(dllimport)
# endif
 
#define CONFIG_FILE "mstecat.csv"
 
#define CLBR_PATTERN 32641
#define MOVETO_DURATION 600
 
NAME32 stsLst[] = {
  "moving",
  "paused",
  "",
  "",
  "calibrating (step1)",
  "calibrating (step2)",
  "calibrating (step3)",
  "",
  "",
  "",
  "",
  "is calibrating",
  "uncalibrated",
  "",
  "",
  "hardware error",
};
 
int sizStsLst = sizeof(stsLst)/sizeof(NAME32);
 
#define MTR_MIN_POSITION -128000000
 
#define MTR_TYPE_UNKNOWN  0x00
#define MTR_TYPE_LINEAR   0x01
#define MTR_TYPE_ROTATION 0x02
 
int getMotorType(char *typeStr)
{
  if (typeStr == NULL) return MTR_TYPE_UNKNOWN;
  if (!stricmp(typeStr,"LINEAR")) return MTR_TYPE_LINEAR;
  if (!stricmp(typeStr,"ROTATION")) return MTR_TYPE_ROTATION;
 return MTR_TYPE_UNKNOWN;
}
 
#define STS_BUFFER_SIZE 20
#define PSAX_BUFFER_SIZE 20

typedef struct
{
  double rbBuffer[PSAX_BUFFER_SIZE];
  double tgtBuffer[PSAX_BUFFER_SIZE];
  int stsBuffer[PSAX_BUFFER_SIZE];
  int grpIdXref[PSAX_BUFFER_SIZE];
  int status[PSAX_BUFFER_SIZE];
  int lnkStsPosPsAx; 
  int lnkIdPosPsAx; 
  int lnkStsStsPsAx; 
  int lnkIdStsPsAx;
  int lnkIdTgtPsAx;
  int lnkStsTgtPsAx;
  int numPsAxGroups;
  int idLastScanned;
} PseudoAxisGroups;

PseudoAxisGroups psAxGrps;

typedef struct
{
  int id;
  float gearRatio;
  MotorGroup grp;
} MotGrpTbl;
#define GRPTBL_SIZE 10
MotGrpTbl grps[GRPTBL_SIZE];
int ngrps = 0;

#define ENABLED_INDEX 0
#define CLBR_INDEX 1
#define IDLE_INDEX 2
#define MOVPOS_INDEX 3
#define MOVNEG_INDEX 4
#define MOVSTS_INDEX 5
#define PCNTMAX_INDEX 6
#define OVERRIDE_INDEX 7
#define ENDPOS_INDEX 8
#define ENDNEG_INDEX 9
#define POS_INDEX 10
#define VEL_INDEX 11
#define AXIS_INDEX 12
#define POSDIF_INDEX 13
#define COUPLED_INDEX 14
#define CNTRL_LOOP_INDEX 15
#define TGTPOS_INDEX 16
#define CLBR_STATE_INDEX 17
#define STATE_INDEX 18
#define NMOT_INDEX 19

// status bit definitions (property "StatusInfo" -> stsBits)
#define STS_IDLE 0x01
#define STS_ENABLED 0x02
#define STS_MOVING 0x04
#define STS_MOVPOS 0x08
#define STS_MOVNEG 0x10
#define STS_ENDPOS 0x20
#define STS_ENDNEG 0x40
#define STS_OVERRIDE 0x80


#define IsCoupledAxisMotor(gid) ((gid)<MST_DYNAGRP_OFFST)

/** PLC status buffer:
Status[1]:= MotorEnabled                        ->    Property needed enable motor    mot.enable [1] 1 parameter int 0 1
Status[2]:= Motor Referenced/Homed;
Status[3]:= Motor Idle;
Status[4]:= Motor MovePositive direction;
Status[5]:= MotorMoveNegative direction
Status[6]:= MotorWorking;
Status[7]:= PEH(percentage of maximum current output);
Status[8]:=Override;
Status[9]:= PosEndSw;
Status[10]:=NegEndSw;
Status[11]:= ActualPosition;
Status[12]:= ActualVelocity;
Status[13]:= AxisState;
Status[14]:= PosDiff  (Soll-Ist Position);
Status[15]:= CoupleState     0-No Coupling, 1 Mater,3 Slave;
Status[16]:= ActiveControlLoopIndex;
Status[17]:= TargetPosition;
Status[18]:= HomingState;
Status[19]:= StateDWord (all info of everything);
Status[20]:= NumMotor;
*/
 
typedef struct
{
  char name[32];
  char units[16];
  int type;
  int grpId;   // -1 if not currently a group member
  int idle;    // bit
  int enabled; // bit
  int moving;  // bit
  int movpos;  // bit
  int movneg;  // bit
  int endpos;  // bit
  int endneg;  // bit
  int axisState; // word ? byte ?
  int coupledState;// word ? byte ?
  int clbrState;// current calibration state in calibration procedure
  int clbrSts;  // is calibrated or not
  int actCntrlLoopIndex;// word ? byte ?
  int microStepsPerStep;
  int mstepsPerCycle;
  float cyclesPerMM;
  float mstepsPerMM;
  float stsBuffer[STS_BUFFER_SIZE];
  double pos;
  double posRb; // readback position
  double tgt;
  double tgtRb; // readback target position
  float overrideSts;
  float lag;  // soll - ist
  float velRb;
  float pcntMaxCur;
  float maxPos; // maximum position (micro steps)
  float minPos; // minimum position (micro steps)
  float offPos; // offset position (micro steps)
  float scalePos; // scale factor (usually -1.0 or 1.0)
  float velMean; // normal velocity (msteps / sec)
  float velMin;
  float velMax;
  float velLwr;
  float velUpr;
  float accUpr;
  float rbTolerance;
  float masterGearRatio; // 1.0 default
  int movCmd;
  int movEtr; // 'estimated time remaining' for move action
  int movPaused;
  int status;
  int cmdSts;
  int lnkStsPos;
  int lnkIdPos; 
  time_t lastCheckedClbr;
} BeckoffMotor;
 
BeckoffMotor *motors = NULL;
 
char gTarget[128];
int gNumMotors = 1;
int gIsInProcess = FALSE;
void cbTgtPos(int id, int cc);
void startUpdateMotor(int mot);
LIB_EXPORT int moveto(int mot,int cmd);
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd);
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp);
LIB_EXPORT int grpmoveto(int grpId,int cmd);
 
typedef struct
{
  char target[128];
  char type[32];
  char units[16];
  int motorNr;
  int mstepsStep;
  int mstepsCycle;
  float velNormal;
  float velMin;
  float velMax;
  float posMax;
  float posMin;
  float posOff;
  float posScale;
  float mstepsMM;
  float rbTolerance;
  float gearRatio;
} MotorInfoStruct;
 
CSVCOLUMN csvBeckoffConfigColumns[] =
{
  { "TARGET","/localhost/cdi",0,CF_TEXT,128,(BYTE *)OFFSETIN(MotorInfoStruct,target),NULL,0 },
  { "TYPE","LINEAR",0,CF_TEXT,32,(BYTE *)OFFSETIN(MotorInfoStruct,type),NULL,0 },
  { "MOTOR","",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,motorNr),NULL,0 },
  { "UNITS","mm",0,CF_TEXT,16,(BYTE *)OFFSETIN(MotorInfoStruct,units),NULL,0 },
  { "MSTEPS_MM","12800",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsMM),NULL,0 },
  { "MSTEPS_CYCLE","200",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsCycle),NULL,0 },
  { "MSTEPS_STEP","64",0,CF_INT32,1,(BYTE *)OFFSETIN(MotorInfoStruct,mstepsStep),NULL,0 },
  { "POSITION_MAXIMUM","768000",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMax),NULL,0 },
  { "POSITION_MINIMUM","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posMin),NULL,0 },
  { "POSITION_OFFSET","0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posOff),NULL,0 },
  { "POSITION_SCALE","1",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,posScale),NULL,0 },
  { "VELOCITY_NORMAL","0.1",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,velNormal),NULL,0 },
  { "VELOCITY_MINIMUM","0.01",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,velMin),NULL,0 },
  { "VELOCITY_MAXIMUM","0.5",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,velMax),NULL,0 },
  { "POSITION_TOLERANCE","0.01",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,rbTolerance),NULL,0 },
  { "GEAR_RATIO","1.0",0,CF_FLOAT,1,(BYTE *)OFFSETIN(MotorInfoStruct,gearRatio),NULL,0 },
};
CSVDB csvBeckoffConfigDb =
{
  csvBeckoffConfigColumns,
  sizeof(csvBeckoffConfigColumns)/sizeof(csvBeckoffConfigColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(MotorInfoStruct)
};
 
NAME64 *mstNames = NULL;
NAME32 *regNames = NULL;
 
int getIntegerFromHardware(char *cmd, int *val)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  dout.dArrayLength = 1;
  dout.dFormat = CF_INT32;
  dout.data.lptr = val;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"RECV",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendIntegerToHardware(char *cmd, int val)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = 1;
  din.dFormat = CF_INT32;
  din.data.lptr = &val;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendIntegerArrayToHardware(char *cmd, int *ival,int length)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = length;
  din.dFormat = CF_INT32;
  din.data.lptr = ival;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendFloatToHardware(char *cmd, float fval)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = 1;
  din.dFormat = CF_FLOAT;
  din.data.fptr = &fval;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendFloatArrayToHardware(char *cmd, float *fval,int length)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = length;
  din.dFormat = CF_FLOAT;
  din.data.fptr = fval;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendDoubleToHardware(char *cmd, double dval)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = 1;
  din.dFormat = CF_DOUBLE;
  din.data.dptr = &dval;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendDoubleArrayToHardware(char *cmd, double *dval,int length)
{
  DTYPE dout, din;
  int cc;
  char dev[128],errstr[64];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
  din.dArrayLength = length;
  din.dFormat = CF_DOUBLE;
  din.data.dptr = dval;
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND",NULL,&din,CA_READ|CA_WRITE,200)) != 0)
  {
    feclog("cannot send %s to hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int getFloatArrayFromHardware(char *cmd, float *vals,int length)
{
  int cc;
  DTYPE dout;
  char dev[128],errstr[64];
  if (vals == NULL) return invalid_parameter;
 
  memset(&dout,0,sizeof(DTYPE));
 
  dout.dArrayLength = length;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = vals;
 
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"RECV",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int getCharArrayFromHardware(char *cmd, char *str,int length)
{
  int cc;
  DTYPE dout;
  char dev[128],errstr[64];
  if (str == NULL) return invalid_parameter;
 
  memset(&dout,0,sizeof(DTYPE));
 
  dout.dArrayLength = length;
  dout.dFormat = CF_TEXT;
  dout.data.cptr = str;
 
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"RECV",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int getFloatFromHardware(char *cmd, float *val)
{
  int cc;
  DTYPE dout;
  char dev[128],errstr[64];
  if (val == NULL) return invalid_parameter;
 
  memset(&dout,0,sizeof(DTYPE));
 
  dout.dArrayLength = 1;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = val;
 
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND.RECV.ATOM",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int getDoubleArrayFromHardware(char *cmd, double *vals,int length)
{
  int cc;
  DTYPE dout;
  char dev[128],errstr[64];
  if (vals == NULL) return invalid_parameter;
 
  memset(&dout,0,sizeof(DTYPE));
 
  dout.dArrayLength = length;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = vals;
 
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"RECV",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int getDoubleFromHardware(char *cmd, double *val)
{
  int cc;
  DTYPE dout;
  char dev[128],errstr[64];
  if (val == NULL) return invalid_parameter;
 
  memset(&dout,0,sizeof(DTYPE));
 
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = val;
 
  sprintf(dev,"%s/%s",gTarget,cmd);
  if ((cc=ExecLinkEx(dev,"SEND.RECV.ATOM",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s from hardware: %s",cmd,GetSystemErrorString((short)cc,errstr));
  }
  return cc;
}
int sendRecvIntegerCmdToHardware(int mot,int *tgtVals,int *rbVals)
{
  DTYPE dout, din;
  char dev[128];
 
  if (mot < 0 || mot > gNumMotors) return illegal_device;
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
 
  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.lptr = tgtVals;
 
  dout.dArrayLength = 6;
  dout.dFormat = CF_INT32;
  dout.data.lptr = rbVals;
 
  sprintf(dev,"%s/%s.CMD",gTarget,motors[mot].name);
  return ExecLink(dev,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE);
}
int sendGroupCmdToHardware(int gid,int *tgtVals)
{
  int cc;
  DTYPE dout, din;
  int rbVals[6];
  char dev[128];
 
  memset(&dout,0,sizeof(DTYPE));
  memset(&din,0,sizeof(DTYPE));
 
  din.dArrayLength = 4;
  din.dFormat = CF_INT32;
  din.data.lptr = tgtVals;
 
  dout.dArrayLength = 6;
  dout.dFormat = CF_INT32;
  dout.data.lptr = rbVals;
 
  sprintf(dev,"%s/CMDALL",gTarget);
  cc = ExecLink(dev,"SEND.RECV.ATOM",&dout,&din,CA_READ|CA_WRITE);
  if (cc != 0) return cc;
  if (rbVals[0] != tgtVals[0]) return hardware_error;
  if (rbVals[1] != 0) return operation_busy;
  if (rbVals[2] != tgtVals[2]) return remitted_data_lost;
  return 0;
}
 
int getRegisteredMotors(NAME64 *names,NAME32 *regnames, int num)
{
  int cc=0;
  DTYPE dout;
  char dev[128],errstr[64];
 
  if (names == NULL || regnames == NULL || num < 1) return invalid_parameter;
 
  memset(&dout,0,sizeof(DTYPE));
  dout.dArrayLength = num;
  dout.dFormat = CF_NAME64;
  dout.data.vptr = (NAME64 *)names;
  sprintf(dev,"%s/NCMOTOR",gTarget);
  if ((cc=ExecLinkEx(dev,"INSTANCES",&dout,NULL,CA_READ,200)) != 0)
  {
    feclog("cannot read %s INSTANCES from CDI: %s",dev,GetSystemErrorString((short)cc,errstr));
  } 
  return cc;
}
int getPsAxGroupIndex(int grpId)
{
  int i;
  for (i=0; i<psAxGrps.numPsAxGroups; i++)
  {
    if (psAxGrps.grpIdXref[i] == grpId) return i;
  }
  return -1;
}
static MotorInfoStruct *motorInfo = NULL;
void initialize(void)
{
  int i, cc, n, nmotorsFromFile=0;
  char dev[64],errstr[64];
 
  if ((cc=csvReadFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb,(void **)&motorInfo)) != 0) 
  {
    feclog("cannot read startup configuration parameters from %s : %s",
      CONFIG_FILE,GetSystemErrorString((short)cc,errstr));
    exit(1);
  }
  nmotorsFromFile = csvBeckoffConfigDb.siz;
 
  if (nmotorsFromFile < 1)
  {
    feclog("no startup configuration parameters from %s given",CONFIG_FILE);
    exit(1);
  }
  strncpy(gTarget,motorInfo[0].target,128);
  gIsInProcess = (strnicmp(gTarget,"/localhost/cdi",14) == 0) ? TRUE : FALSE;
 
  if (getIntegerFromHardware("NumMotor",&gNumMotors) != 0) { gNumMotors = nmotorsFromFile; } //exit(1);
  if ((motors=(BeckoffMotor *)calloc(gNumMotors,sizeof(BeckoffMotor))) == NULL)
  {
    feclog("cannot allocate space for %d motors",gNumMotors);
    exit(1);
  }
  if ((mstNames=(NAME64 *)calloc(gNumMotors,sizeof(NAME64))) == NULL)
  {
    feclog("cannot allocate motor name list ! (aborting ...)");
    exit(1);
  }
  if ((regNames=(NAME32 *)calloc(gNumMotors,sizeof(NAME32))) == NULL)
  {
    feclog("cannot allocate motor register name list ! (aborting ...)");
    exit(1);
  }
  if ((cc=getRegisteredMotors(mstNames,regNames,gNumMotors)) != 0)
  {
    feclog("cannot read template instances from CDI ! (aborting ...)");
    exit(1);
  }
 
  for (i=0; i<gNumMotors; i++)
  {
    if (strlen(mstNames[i].name) == 0) break;
    strncpy(motors[i].name,mstNames[i].name,32);
    sprintf(dev,"%s.mSteps",motors[i].name);
    getIntegerFromHardware(dev,&motors[i].microStepsPerStep);
    sprintf(dev,"%s.fSteps",motors[i].name);
    getIntegerFromHardware(dev,&motors[i].mstepsPerCycle);
    motors[i].mstepsPerCycle *= motors[i].microStepsPerStep; // ??
    for (n=0; n<nmotorsFromFile; n++)
    {
      if (motorInfo[n].motorNr == i+1)
      {
        strncpy(motors[i].units,motorInfo[n].units,16);
        motors[i].maxPos = motorInfo[n].posMax;
        motors[i].minPos = motorInfo[n].posMin;
        motors[i].offPos = motorInfo[n].posOff;
        motors[i].scalePos = motorInfo[n].posScale;
        motors[i].velMean = motorInfo[n].velNormal;
        motors[i].microStepsPerStep = motorInfo[n].mstepsStep;
        motors[i].mstepsPerCycle = motorInfo[n].mstepsCycle;
        motors[i].mstepsPerMM = motorInfo[n].mstepsMM;
        motors[i].cyclesPerMM = motorInfo[n].mstepsMM/motorInfo[n].mstepsCycle;
        motors[i].type = getMotorType(motorInfo[n].type);
        motors[i].rbTolerance = motorInfo[n].rbTolerance;
        motors[i].velLwr = motorInfo[n].velMin;
        motors[i].velUpr = motorInfo[n].velMax;
        motors[i].masterGearRatio = motorInfo[n].gearRatio;
        break;
      }
    }
    startUpdateMotor(i);
  }
  if (i < gNumMotors)
  {
    feclog("hardware claimed %d motors, but only %d configured",gNumMotors,i);
    gNumMotors = i;
  }

  psAxGrps.idLastScanned = -1; /* unknown at server startup */
}
int SimulateData(int start,int range)
{
  /* This routine is useful for test servers  */
  return start + rand()%(range);
}
void cbSts(int id, int cc)
{
  int i;
  MotorGroup grp;
  char dev[128];

  if (id < 0 || id >= gNumMotors) return; // huh?
  motors[id].lnkStsPos = cc;
  
  if (cc != 0)
  {
    motors[id].status = MTR_ERR;
    return;
  }
  motors[id].status &= ~MTR_ERR;
  motors[id].moving = (int)motors[id].stsBuffer[MOVSTS_INDEX];
  if (motors[id].moving == 0) motors[id].status &= ~MTR_MOVE; else motors[id].status |= MTR_MOVE;
 
  motors[id].pos = motors[id].posRb = motors[id].stsBuffer[POS_INDEX];
  motors[id].tgtRb = motors[id].stsBuffer[TGTPOS_INDEX];
 
  if (motors[id].stsBuffer[CLBR_INDEX] == 1) 
  {
    motors[id].status &= ~MTR_UNCLBR;
    motors[id].status &= ~MTR_ISCLBR;
    //motors[id].status |= MTR_CLBR;
  }
  motors[id].idle = (int)motors[id].stsBuffer[IDLE_INDEX];
  motors[id].enabled = (int)motors[id].stsBuffer[ENABLED_INDEX];
  motors[id].lag = motors[id].stsBuffer[POSDIF_INDEX];
  motors[id].velRb = motors[id].stsBuffer[VEL_INDEX];
  motors[id].pcntMaxCur = motors[id].stsBuffer[PCNTMAX_INDEX];
  motors[id].overrideSts = motors[id].stsBuffer[OVERRIDE_INDEX];

  motors[id].movpos = (int)motors[id].stsBuffer[MOVPOS_INDEX];
  motors[id].movneg = (int)motors[id].stsBuffer[MOVNEG_INDEX];
  motors[id].endpos = (int)motors[id].stsBuffer[ENDPOS_INDEX];
  motors[id].endneg = (int)motors[id].stsBuffer[ENDNEG_INDEX];
  motors[id].axisState = (int)motors[id].stsBuffer[AXIS_INDEX];
  motors[id].coupledState = (int)motors[id].stsBuffer[COUPLED_INDEX];
  motors[id].clbrState = (int)motors[id].stsBuffer[CLBR_STATE_INDEX];
  motors[id].clbrSts = (int)motors[id].stsBuffer[CLBR_INDEX];
  motors[id].actCntrlLoopIndex = (int)motors[id].stsBuffer[CNTRL_LOOP_INDEX];

  if (fabs(motors[id].pos-motors[id].tgt) < motors[id].rbTolerance && motors[id].moving == 0)
  {
    if (motors[id].movCmd != PROC_STS_IDLE) 
    {
      if (motors[id].grpId >= 0)
      { // is a group member
        if (motors[id].coupledState == 1 && // is the group master
            getGrpMembers(motors[id].grpId,&grp) == 0) // filled out the grp descriptor
        { // master is always the first member in the group
          for (i=id+1; i<gNumMotors; i++)
          { // decouple all coupled motors from axis
            if (IsMotorInGroup(i,&grp))
            { // auto decouple the slaves ...
              sprintf(dev,"%s.DecoupleAxis",motors[i].name);
              if ((cc=sendFloatToHardware(dev,1)) != 0)
              {
                motors[i].status |= MTR_ERR;
                continue;
              }
            }
          }
          grpmoveto(motors[id].grpId,PROC_STS_IDLE); // set this automatically
        }
      }
      else
      { // moving as single motor 
        moveto(id,PROC_STS_IDLE);
      }
    }
  }
  if (motors[id].tgtRb != motors[id].tgt)
  {
    if (motors[id].coupledState != 3)
    { // not a coupled slave !
      feclog("Unexpected jump in target position ! (%g to %g)",
        motors[id].tgt,motors[id].tgtRb);
    }
    motors[id].tgt = motors[id].tgtRb;
  }
  
  //TODO: finish
}
void cbPosPsAx(int id, int cc)
{
  psAxGrps.lnkStsPosPsAx = cc;
}
void cbTgtPsAx(int id, int cc)
{
  psAxGrps.lnkStsTgtPsAx = cc;
}
void cbStsPsAx(int id, int cc)
{
  int i;
  psAxGrps.lnkStsStsPsAx = cc;
  if (cc == 0)
  {
    for (i=0; i< psAxGrps.numPsAxGroups; i++)
    { //1:moving,2:idle,3:Error
      if (psAxGrps.stsBuffer[i] == 0) psAxGrps.status[i] = MTR_IDLE;
      if (psAxGrps.stsBuffer[i] == 1) psAxGrps.status[i] = MTR_MOVE;
      if (psAxGrps.stsBuffer[i] == 2) psAxGrps.status[i] = MTR_IDLE;
      if (psAxGrps.stsBuffer[i] == 3) psAxGrps.status[i] = MTR_ERR;
    }
  }
}
void cbPos(int id, int cc)
{
  static int zeroCounter = 0;
  if (id < 0 || id >= gNumMotors) return; // huh?
  motors[id].lnkStsPos = cc;
  if (motors[id].status & MTR_MOVE)
  { /* motor is moving */
    if (motors[id].posRb == 0 && motors[id].tgt != 0)
    { /* beckoff gives a spurious '0' reading from time to time */
      if (zeroCounter++ < 2) return;
    }
  }
  zeroCounter = 0;
  motors[id].pos = motors[id].posRb;
  if (motors[id].pos == motors[id].tgt)
  {
    if (motors[id].movCmd != PROC_STS_IDLE) moveto(id,PROC_STS_IDLE);
  }
}
#define MTR_UPDATE_INTERVAL 200
 
void startUpdateMotor(int mot)
{
  DTYPE dout;
  char dev[128];
  if (mot < 0 || mot > gNumMotors) return;
 
  memset(&dout,0,sizeof(DTYPE));
 
  dout.dArrayLength = STS_BUFFER_SIZE;
  dout.dFormat = CF_FLOAT;
  dout.data.fptr = motors[mot].stsBuffer;
 
  sprintf(dev,"%s/%s.Status",gTarget,motors[mot].name);
  motors[mot].lnkIdPos = AttachLinkEx(dev,"RECV",&dout,NULL,CA_READ,MTR_UPDATE_INTERVAL,cbSts,CM_TIMER,mot);

  dout.dArrayLength = PSAX_BUFFER_SIZE;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = psAxGrps.rbBuffer;

  sprintf(dev,"%s/PsAxPos",gTarget);
  psAxGrps.lnkIdPosPsAx = AttachLink(dev,"RECV",&dout,NULL,CA_READ,MTR_UPDATE_INTERVAL,cbPosPsAx,CM_TIMER);

  dout.dArrayLength = PSAX_BUFFER_SIZE;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = psAxGrps.tgtBuffer;

  sprintf(dev,"%s/GetPsAxTgtPos",gTarget);
  psAxGrps.lnkIdTgtPsAx = AttachLink(dev,"RECV",&dout,NULL,CA_READ,MTR_UPDATE_INTERVAL,cbTgtPsAx,CM_TIMER);

  dout.dArrayLength = PSAX_BUFFER_SIZE;
  dout.dFormat = CF_INT32;
  dout.data.lptr = psAxGrps.stsBuffer;
  sprintf(dev,"%s/PsAxStat",gTarget);
  psAxGrps.lnkIdStsPsAx = AttachLink(dev,"RECV",&dout,NULL,CA_READ,MTR_UPDATE_INTERVAL,cbStsPsAx,CM_TIMER);
}
 
LIB_EXPORT int getUnits(int mot,char *unitsBuffer,int bufferLength)
{
  int len;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  len = MIN(bufferLength,16);
  strncpy(unitsBuffer,motors[mot].units,len);
  return 0;
}
LIB_EXPORT int getPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].pos * motors[mot].scalePos + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getGrpPos(int grpId,double *val)
{
  int i, m;
  if (val == NULL) return invalid_parameter;
  if (IsCoupledAxisMotor(grpId))
  {
    for (i=0; i<psAxGrps.numPsAxGroups; i++)
    {
      if (psAxGrps.grpIdXref[i] == grpId)
      { // got it
        *val = psAxGrps.rbBuffer[i];
        return 0;
      }
    }
  }
  else
  {
    for (i=0; i<ngrps; i++)
    {
      if (grps[i].id != grpId) continue;
      m = grps[i].grp.grphead - 1;
      *val = motors[m].pos * motors[m].scalePos + motors[m].offPos;
      return 0;
    }
  }
  return invalid_index;
}
LIB_EXPORT int getPosMin(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].minPos * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getPosMax(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].maxPos * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int getPosOff(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].offPos;
  return 0;
}
LIB_EXPORT int setPosOff(int mot,double val)
{
  int cc;
  char errstr[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  motors[mot].offPos = (float)val;
  // save and restore:
  motorInfo[mot].posOff = motors[mot].offPos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getPosScale(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = (double)motors[mot].scalePos;
  return 0;
}
LIB_EXPORT int setPosScale(int mot,double val)
{
  int cc;
  char errstr[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  motors[mot].scalePos = (float)val;
  // save and restore:
  motorInfo[mot].posScale = motors[mot].scalePos;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getVel(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].velMean;
  return 0;
}
//TODO: trap velocity > upr limit or < lwr limit with software ?
LIB_EXPORT int setVel(int mot,double val)
{
  int cc;
  double dvals[2];
  char dev[64];
  char errstr[64];
 
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (val < motors[mot].velLwr ||
      val > motors[mot].velUpr) return out_of_range;
  sprintf(dev,"%s.Velocity",motors[mot].name);
  dvals[0] = 1;
  dvals[1] = val;
  cc = sendDoubleArrayToHardware(dev,dvals,2);
  if (cc != 0) return cc;
  motors[mot].velMean = (float)(val);
  motorInfo[mot].velNormal = motors[mot].velMean;
  if ((cc=csvWriteFile(NULL,CONFIG_FILE,&csvBeckoffConfigDb)) != 0) 
  {
    feclog("could not update configuration file! : %s",GetSystemErrorString((short)cc,errstr));
  }
  return 0;
}
LIB_EXPORT int getVelLwr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].velLwr;
  return 0;
}
LIB_EXPORT int getVelUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].velUpr;
  return 0;
}
LIB_EXPORT int getAccUpr(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].accUpr;
  return 0;
}
 
// TODO:
LIB_EXPORT int getAcc(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val =  SimulateData(5,2);
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int mstepsPerMM(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].mstepsPerMM;
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int mstepsPerStep(int mot,int *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].microStepsPerStep;
  return 0;
}
//TODO: necessary ?
LIB_EXPORT int stepsPerRadian(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *val = motors[mot].mstepsPerCycle/6.2832;
  return 0;
}
 
int sendMoveCmd(int mot,int useEndSwitch)
{
  char dev[64];
  sprintf(dev,"%.32s.MoveAbs",motors[mot].name);
  return motors[mot].cmdSts = sendDoubleToHardware(dev,1);
}
int stopMoveCmd(int mot)
{
  char dev[64];
  sprintf(dev,"%.32s.Stop",motors[mot].name);
  motors[mot].cmdSts = sendIntegerToHardware(dev,1);
  return 0;
}
//TODO: fix
int sendGrpMoveCmd(int gid,int useEndSwitch)
{
  int tgtVals[4];
  tgtVals[0] = 1; // load target position
  tgtVals[1] = 1; // ??
  tgtVals[2] = useEndSwitch ? 0 : 1;// with END Switch control
  tgtVals[3] = 0;
  return sendGroupCmdToHardware(gid,tgtVals);
}
//TODO: fix
int stopGrpMoveCmd(int gid)
{
  int tgtVals[4];
  tgtVals[0] = 2;
  tgtVals[1] = 1;
  tgtVals[2] = 0;
  tgtVals[3] = 0;
  return sendGroupCmdToHardware(gid,tgtVals);
}

//TODO: fix
LIB_EXPORT int scanprg(int mot,int *prg)
{
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (prg == NULL) return invalid_parameter;
  *prg = motors[mot].movEtr;
  return 0;
}
#define SCANLENGTH 5000
float scanbuffer[SCANLENGTH*2+4];
int scanlength = SCANLENGTH*2 + 1;
LIB_EXPORT int scansts(int mot,DTYPE *dout)
{
  char dev[64];
  int i, dlen, cc;
  float *vals;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (dout == NULL) return invalid_parameter;
  if (dout->dArrayLength == 1)
  {
    return PutValuesFromLong(dout,&motors[mot].movCmd,1);
  }
  if (dout->dFormat != CF_FLOAT) return illegal_format;
  sprintf(dev,"%.32s.ScanArray",motors[mot].name);
  cc = getFloatArrayFromHardware(dev,scanbuffer,scanlength);
  if (cc) return cc;
  // 1st element contains 1 + 2 x array length
  dlen = (int)scanbuffer[0];
  dlen--;  // subtract the 1st element
  if (dlen < 0) return invalid_data;
  vals = dout->data.fptr;
  if (dlen > (int)dout->dArrayLength) dlen = (int)dout->dArrayLength;
  dout->dArrayLength = dlen;
  for (i=0; i<dlen; i++)
  {
    vals[i] = scanbuffer[1+i];
  }
  return 0;
}
LIB_EXPORT int scanto(int mot,int cmd,DTYPE *din)
{
  double etr;
  int cc,sts;
  char dev[64];
  static double dvals[32];
  static int dlen = 0;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  if (!motors[mot].enabled) return device_offline;
  sprintf(dev,"%.32s.StepScan",motors[mot].name);
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      dlen = MIN(din->dArrayLength,32);
      if ((cc=GetValuesAsDouble(din,dvals,32)) != 0) return cc;
      if (sendDoubleArrayToHardware(dev,dvals,dlen) == 0)
      {
        //TODO: fix this
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      break;
    case PROC_STS_RESUME:
      if (sendDoubleArrayToHardware(dev,dvals,dlen) == 0)
      {
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
  }
  return 0;
}

LIB_EXPORT int movests(int mot,int *sts)
{
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (sts == NULL) return invalid_parameter;
  *sts = motors[mot].movCmd;
  return 0;
}
LIB_EXPORT int moveprg(int mot,int *prg)
{
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (prg == NULL) return invalid_parameter;
  *prg = motors[mot].movEtr;
  return 0;
}
LIB_EXPORT int moveto(int mot,int cmd)
{
  double etr;
  int sts;
  char dev[64];
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  if (!motors[mot].enabled) return device_offline;
  sprintf(dev,"%.32s.MoveAbs",motors[mot].name);
  switch (cmd)
  {
    case PROC_STS_IDLE:
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendDoubleToHardware(dev,motors[mot].tgt) == 0)
      {
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        //etr /= (double)(motors[mot].velRps * motors[mot].mstepsPerCycle);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    case PROC_STS_PAUSE:
      stopMoveCmd(mot);
      motors[mot].movPaused = TRUE;
      motors[mot].status |= MTR_PAUSE;
      break;
    case PROC_STS_RESUME:
      if (sendDoubleToHardware(dev,motors[mot].tgt) == 0)
      {
        motors[mot].movPaused = FALSE;
        motors[mot].status &= ~MTR_PAUSE;
      }
      break;
    case PROC_STS_STOP:
      stopMoveCmd(mot);
      motors[mot].movEtr = 0;
      motors[mot].movPaused = FALSE;
      motors[mot].status = MTR_IDLE;
      break;
  }
  return 0;
}
//TODO: test ?
LIB_EXPORT int movetoNoEndSwitch(int mot,int cmd)
{
  double etr;
  int sts;
  motors[mot].movCmd = cmd;
  sts = motors[mot].status & MTR_CLBR;
  if ((motors[mot].status & MTR_CLBR) != 0) return not_allowed;
  if (!motors[mot].enabled) return device_offline;
  switch (cmd)
  {
    case PROC_STS_START:
      if (motors[mot].status & MTR_MOVE) return operation_busy;
      if (sendMoveCmd(mot,FALSE) == 0)
      {
        etr = fabs(motors[mot].tgt - motors[mot].pos);
        etr /= (double)(motors[mot].velMean);
        motors[mot].movEtr = (int)etr;
        motors[mot].movPaused = FALSE;
        motors[mot].status |= MTR_MOVE;
      }
      else
      {
        motors[mot].status |= MTR_ERR;
      }
      break;
    default:
      return moveto(mot,cmd);
  }
  return 0;
}
int getGrpTblIndex(MotorGroup *grp) 
{ 
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (memcmp(&grps[i].grp,grp,sizeof(MotorGroup))) continue;
    return i;
  }
  return -1; 
}
int addToGrpTbl(MotorGroup *grp) 
{ 
  int i = getGrpTblIndex(grp);
  if (i >= 0) return i;
  if (ngrps < GRPTBL_SIZE)
  {
    i = ngrps;
    grps[i].gearRatio = 0;
    memcpy(&grps[i].grp,grp,sizeof(MotorGroup));
    ngrps++;
    return i;
  }
  return -1; 
}
LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp)
{
  int i;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    memcpy(grp,&grps[i].grp,sizeof(MotorGroup));
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int setGrpMembers(int grpId,MotorGroup *grp)
{
  int g;
  if (grpId < MST_DYNAGRP_OFFST)
  {
    if ((g=getPsAxGroupIndex(grpId)) < 0)
    {
      if (psAxGrps.numPsAxGroups >= PSAX_BUFFER_SIZE) return resources_exhausted;
      g = psAxGrps.numPsAxGroups;
      psAxGrps.numPsAxGroups++;
    }
    psAxGrps.grpIdXref[g] = grpId;
  }
  else
  {
    if ((g=addToGrpTbl(grp)) < 0) return invalid_index;
    grps[g].id = grpId;
  }
//  memcpy(&grps[g].grp,grp,sizeof(MotorGroup));
  return 0;
}
LIB_EXPORT int setGrpGearRatio(int grpId,double gearRatio)
{
  int i;
  if (gearRatio == 0) return out_of_range;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    grps[i].gearRatio = (float)gearRatio;
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int getGrpGearRatio(int grpId,double *gearRatio)
{
  int i;
  if (gearRatio == NULL) return invalid_parameter;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    *gearRatio = (double)grps[i].gearRatio;
    return 0;
  }
  return invalid_index;
}
//TODO: fix
LIB_EXPORT int grpmoveto(int grpId,int cmd)
{
  int cc=0, i, g=-1;
  MotorGroup grp;
  double etr, gearRatio=0;
  float fvals[4];
  int ivals[2];
  char dev[128];

  if (IsCoupledAxisMotor(grpId))
  { //TODO GROUP
    if ((i=getPsAxGroupIndex(grpId)) < 0) return invalid_index;
    switch (cmd)
    {
      case PROC_STS_IDLE:
        psAxGrps.status[i] = MTR_IDLE;
        break;
      case PROC_STS_RESUME:
        psAxGrps.status[i] &= ~MTR_PAUSE;
      case PROC_STS_START:
        if (psAxGrps.status[i] & MTR_MOVE) return operation_busy;
        ivals[0] = grpId;
        ivals[1] = 1; // start move
        cc = sendIntegerArrayToHardware("PsAxCtrl",ivals,2);
        psAxGrps.status[i] |= MTR_MOVE;
        break;
      case PROC_STS_PAUSE:
        psAxGrps.status[i] |= MTR_PAUSE;
      case PROC_STS_STOP:
        ivals[0] = grpId;
        ivals[1] = 2; // stop move
        cc = sendIntegerArrayToHardware("PsAxCtrl",ivals,2);
        if (cmd == PROC_STS_STOP) psAxGrps.status[i] = MTR_IDLE;
        break;
    }
    return cc;
  }
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  if (grp.grphead > 0) 
  { // group master defined from above
    g = grp.grphead - 1;
  }
  else
  { // not defined: take the first in the list
    for (i=0; i<gNumMotors; i++) if (IsMotorInGroup(i,&grp)) break;
    if (i == gNumMotors) return configuration_error;
    g = i;
  }
  if (cmd == PROC_STS_START)
  { // first things first
    for (i=0; i<gNumMotors; i++)
    { 
      if (IsMotorInGroup(i,&grp))
      { // don't allow if anybody is being calibrated or is moving !
        if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
        if ((motors[i].status & MTR_MOVE) != 0) return operation_busy;
        if (g == i) 
        { // fix the coupling parameters 
          getGrpGearRatio(grpId,&gearRatio);
          if (gearRatio == 0) gearRatio = (double)motors[i].masterGearRatio;
          fvals[0] = 1;
          fvals[1] = (float)(g + 1);
          fvals[2] = (float)gearRatio;
        }
      }
    }
    for (i=0; i<gNumMotors; i++)
    { // couple all motors in group to same axis
      if (IsMotorInGroup(i,&grp))
      { // this one is a group member !      
        motors[i].grpId = grpId; // make a note of the group id
        if (g == i) continue; // don't couple the master
        sprintf(dev,"%s.CoupleAxis",motors[i].name);
        if ((cc=sendFloatArrayToHardware(dev,fvals,3)) != 0)
        { // couple all other members to the master
          motors[i].status |= MTR_ERR;
          continue;
        }
      }
    }
  }
  for (i=0; i<gNumMotors; i++)
  { // send move commands only to the master
    if (IsMotorInGroup(i,&grp))
    {       
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          motors[i].grpId = -1;
          break;
        case PROC_STS_START:
          if (i == g) moveto(g,cmd);
          etr = fabs(motors[i].tgt - motors[i].pos);
          etr /= (double)(motors[i].velMean);
          //etr /= (double)(motors[i].velRps * motors[i].mstepsPerCycle);
          motors[i].movEtr = (int)etr;
          motors[i].movPaused = FALSE;
          motors[i].status |= MTR_MOVE;
          break;
        case PROC_STS_PAUSE:
          if (i == g) stopMoveCmd(g);
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_RESUME:
          if (i == g) moveto(g,cmd);
          motors[i].movPaused = FALSE;
          motors[i].status &= ~MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          if (i == g) stopMoveCmd(g);
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
      }
    }
  }
  if (cmd == PROC_STS_STOP)
  { // finished
    for (i=0; i<gNumMotors; i++)
    { // decouple all coupled motors from axis
      if (IsMotorInGroup(i,&grp))
      {       
        if (i == g) continue;
        sprintf(dev,"%s.DecoupleAxis",motors[i].name);
        if ((cc=sendFloatToHardware(dev,1)) != 0)
        {
          motors[i].status |= MTR_ERR;
          continue;
        }
      }
    }
  }
  return cc;
}
LIB_EXPORT int grpscansts(int grpId,DTYPE *dout)
{
  char dev[64];
  int i, mot, dlen, cc;
  float *vals;
  MotorGroup grp;
  //TODO GROUP ???
  if (IsCoupledAxisMotor(grpId))
  { //TODO GROUP
    if ((i=getPsAxGroupIndex(grpId)) < 0) return invalid_index;
    if (psAxGrps.idLastScanned >= 0 && 
        psAxGrps.idLastScanned != grpId) return invalid_data;
    strcpy(dev,"psAxScanArray");
    goto getScanData;
  }
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors && !IsMotorInGroup(i,&grp); i++);
  if (i == gNumMotors) return illegal_device;
  mot = i;
  if (dout == NULL) return invalid_parameter;
  if (dout->dArrayLength == 1)
  {
    return PutValuesFromLong(dout,&motors[mot].movCmd,1);
  }
  if (dout->dFormat != CF_FLOAT) return illegal_format;
  sprintf(dev,"%.32s.ScanArray",motors[mot].name);
getScanData:
  cc = getFloatArrayFromHardware(dev,scanbuffer,scanlength);
  if (cc) return cc;
  dlen = (int)scanbuffer[0]; dlen--;
  vals = dout->data.fptr;
  if (dlen > (int)dout->dArrayLength) dlen = (int)dout->dArrayLength;
  dout->dArrayLength = dlen;
  dlen /= 2;
  for (i=0; i<dlen; i++)
  {
    vals[i] = scanbuffer[1+i];
    vals[i+dlen] = scanbuffer[1+dlen+i];
  }
  return 0;
}
LIB_EXPORT int grpscanto(int grpId,int cmd,DTYPE *din)
{
  int cc, i, g=-1;
  MotorGroup grp;
  double etr, gearRatio;
  float fvals[4];
  int ivals[6];
  char dev[128];
  static double dvals[32];
  static int dlen = 0;

  if (IsCoupledAxisMotor(grpId))
  { //TODO GROUP
    if ((i=getPsAxGroupIndex(grpId)) < 0) return invalid_index;
    switch (cmd)
    {
      case PROC_STS_IDLE:
        psAxGrps.status[i] = MTR_IDLE;
        break;
      case PROC_STS_RESUME:
        psAxGrps.status[i] &= ~MTR_PAUSE;
      case PROC_STS_START:
        if (psAxGrps.status[i] & MTR_MOVE) return operation_busy;
   	    dlen = MIN(din->dArrayLength,32);
        if ((cc=GetValuesAsDouble(din,dvals,32)) != 0) return cc;
        dvals[0] += grpId;
        if (sendDoubleArrayToHardware("PsAxStepScan",dvals,dlen) == 0)
        {
		      psAxGrps.status[i] |= MTR_MOVE;
          psAxGrps.idLastScanned = grpId;
  	    }
        else
        {
          psAxGrps.status[i] |= MTR_ERR;
        }
        psAxGrps.status[i] |= MTR_MOVE;
        break;
      case PROC_STS_PAUSE:
        psAxGrps.status[i] |= MTR_PAUSE;
      case PROC_STS_STOP:
        ivals[0] = grpId;
        ivals[1] = 2; // stop move
        cc = sendIntegerArrayToHardware("PsAxCtrl",ivals,2);
        if (cmd == PROC_STS_STOP) psAxGrps.status[i] = MTR_IDLE;
        break;
	}
    return cc;
  }
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  if (grp.grphead > 0) 
  { // group master defined from above
    g = grp.grphead - 1;
  }
  else
  { // not defined: take the first in the list
    for (i=0; i<gNumMotors; i++) if (IsMotorInGroup(i,&grp)) break;
    if (i == gNumMotors) return configuration_error;
    g = i;
  }
  if (cmd == PROC_STS_START)
  { // first things first
    for (i=0; i<gNumMotors; i++)
    { 
      if (IsMotorInGroup(i,&grp))
      { // don't allow if anybody is being calibrated or is moving !
        if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
        if ((motors[i].status & MTR_MOVE) != 0) return operation_busy;
        if (g == i) 
        { // fix the coupling parameters 
          getGrpGearRatio(grpId,&gearRatio);
          if (gearRatio == 0) gearRatio = (double)motors[i].masterGearRatio;
          fvals[0] = 1;
          fvals[1] = (float)(g + 1);
          fvals[2] = (float)gearRatio;
        }
      }
    }
    for (i=0; i<gNumMotors; i++)
    { // couple all motors in group to same axis
      if (IsMotorInGroup(i,&grp))
      { // this one is a group member !      
        motors[i].grpId = grpId; // make a note of the group id
        if (g == i) continue; // don't couple the master
        sprintf(dev,"%s.CoupleAxis",motors[i].name);
        if ((cc=sendFloatArrayToHardware(dev,fvals,3)) != 0)
        { // couple all other members to the master
          motors[i].status |= MTR_ERR;
          continue;
        }
      }
    }
  }  
  for (i=0; i<gNumMotors; i++)
  { // send move commands only to the master
    if (IsMotorInGroup(i,&grp))
    {       
      motors[i].movCmd = cmd;
      switch (cmd)
      {
        case PROC_STS_IDLE:
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
        case PROC_STS_START:
          if (i == g) scanto(g,cmd,din);
          etr = fabs(motors[i].tgt - motors[i].pos);
          etr /= (double)(motors[i].velMean);
          motors[i].movEtr = (int)etr;
          motors[i].movPaused = FALSE;
          motors[i].status |= MTR_MOVE;
          break;
        case PROC_STS_PAUSE:
          if (i == g) stopMoveCmd(g);
          motors[i].movPaused = TRUE;
          motors[i].status |= MTR_PAUSE;
          break;
        case PROC_STS_RESUME:
          if (i == g) scanto(g,cmd,din);
          motors[i].movPaused = FALSE;
          motors[i].status &= ~MTR_PAUSE;
          break;
        case PROC_STS_STOP:
          if (i == g) stopMoveCmd(g);
          motors[i].movEtr = 0;
          motors[i].movPaused = FALSE;
          motors[i].status = MTR_IDLE;
          break;
      }
    }
  }
  if (cmd == PROC_STS_STOP)
  { // finished
    for (i=0; i<gNumMotors; i++)
    { // decouple all coupled motors from axis
      if (IsMotorInGroup(i,&grp))
      {       
        if (i == g) continue;
        sprintf(dev,"%s.DecoupleAxis",motors[i].name);
        if ((cc=sendFloatToHardware(dev,1)) != 0)
        {
          motors[i].status |= MTR_ERR;
          continue;
        }
      }
    }
  }
  return 0;
}
LIB_EXPORT int setPos(int mot,double val)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (val < motors[mot].minPos || val > motors[mot].maxPos) 
  {
    if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
  }
  //TODO: fill this in !
  return 0;
}
LIB_EXPORT int setTgtPos(int mot,double val)
{
  if (mot < 0 || mot >= gNumMotors) return illegal_device;
  if (motors[mot].status & MTR_MOVE) return operation_busy;
  if (!motors[mot].enabled) return device_offline;
  val -= motors[mot].offPos;
  if (motors[mot].scalePos != 0) val /= motors[mot].scalePos;
  if (val < motors[mot].minPos || val > motors[mot].maxPos) 
  {
    if ((motors[mot].status & MTR_ISCLBR) == 0) return out_of_range;
  }
  motors[mot].tgt = val;  
  return 0;
}
LIB_EXPORT int getTgtPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
 
  *val = motors[mot].tgt * motors[mot].scalePos  + motors[mot].offPos;
  return 0;
}
LIB_EXPORT int setGrpTgtPos(int grpId,double val)
{
  int cc, i, g=-1;
  double dvals[4];
  MotorGroup grp;

  if (IsCoupledAxisMotor(grpId))
  { //TODO GROUP
    if ((i=getPsAxGroupIndex(grpId)) < 0) return invalid_index;
    dvals[0] = grpId;
    dvals[1] = val;
    if ((cc=sendDoubleArrayToHardware("SetPsAxTgtPos",dvals,2)) != 0) return cc;
    psAxGrps.tgtBuffer[i] = (float)val;
    return 0;
  }
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors; i++)
  { // don't allow if anybody is being calibrated !
    if ((motors[i].status & MTR_CLBR) != 0) return not_allowed;
  }
  for (i=0; i<gNumMotors && g < 0; i++) if (IsMotorInGroup(i,&grp)) g = i;
  return setTgtPos(g,val);
}
LIB_EXPORT int getGrpTgtPos(int grpId,double *val)
{
  int cc, i, g=-1;
  MotorGroup grp;

  if (val == NULL) return invalid_parameter;
  if (IsCoupledAxisMotor(grpId))
  { //TODO GROUP
    if ((i=getPsAxGroupIndex(grpId)) < 0) return invalid_index;
    *val = psAxGrps.tgtBuffer[i];
    return 0;
  }
  if ((cc=getGrpMembers(grpId,&grp)) != 0) return cc;
  for (i=0; i<gNumMotors && g < 0; i++) if (IsMotorInGroup(i,&grp)) g = i;
  if (g < 0 || g > gNumMotors) return illegal_device;
  return getTgtPos(g,val);
}
//TODO GROUP ???
LIB_EXPORT int calibr(int mot,int *sts)
{
  int cc = 0;
  char dev[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if ((motors[mot].status & 0xff) != MTR_IDLE) return not_allowed;
  if (!motors[mot].enabled) return device_offline;
  if (motors[mot].type == MTR_TYPE_LINEAR)
  { // motor has end switch (n.b. setTgtPosMs() will reverse the scale and offset)
    motors[mot].status |= MTR_ISCLBR;
    sprintf(dev,"%.32s.Home",motors[mot].name);
    motors[mot].cmdSts = sendIntegerToHardware(dev,1);
  }
  else
  { // motor has no end switch
    motors[mot].status |= MTR_ISCLBR;
    sprintf(dev,"%.32s.Home",motors[mot].name);
    motors[mot].cmdSts = sendIntegerToHardware(dev,1);
  }
  if (sts != NULL) *sts = 0;
  return 0;
}
//TODO: check
LIB_EXPORT int calibrToOffset(int mot,double val)
{
  int cc;
  char dev[64];
 
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (motors[mot].type != MTR_TYPE_LINEAR) return not_allowed;
  motors[mot].status |= MTR_ISCLBR;
  sprintf(dev,"%.32s.Home",motors[mot].name);
  cc = motors[mot].cmdSts = sendDoubleToHardware(dev,val);
  return cc;
}
LIB_EXPORT int getonline(int mot,int *sts)
{
  if (mot < 0 || mot > gNumMotors) return illegal_device; 
  if (sts == NULL) return argument_list_error;
  *sts = motors[mot].enabled;
  return 0;
}
LIB_EXPORT int setonline(int mot,int sts)
{
  int cc, val;
  char dev[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  sprintf(dev,"%.32s.Enable",motors[mot].name);
  val = sts ? 1 : 0;
  cc = motors[mot].cmdSts = sendIntegerToHardware(dev,val);
  return cc;
}

LIB_EXPORT int reset(int mot,int *sts)
{
  char dev[64];
  if (mot < 0 || mot > gNumMotors) return illegal_device; 
  motors[mot].status = MTR_UNCLBR;
  if (sts != NULL) *sts = 0;
  sprintf(dev,"%.32s.Reset",motors[mot].name);
  motors[mot].cmdSts = sendIntegerToHardware(dev,1);
  //calibr(mot,sts);
 
  return 0;
}
int chkClbrFlag(int mot)
{
  int vals[4], rbvals[6];
 
  vals[0] = 40; vals[1] = 1; vals[2] = 15; vals[3] = 0;
  motors[mot].cmdSts = sendRecvIntegerCmdToHardware(mot,vals,rbvals);
 
  if (rbvals[3] == CLBR_PATTERN) return TRUE;
  motors[mot].status |= MTR_UNCLBR;
  return FALSE;
}
//TODO: necessary ?
int chkEndStation(int mot,int mask)
{
//char dev[128];
  int reg = 0;
//sprintf(dev,"%s.regExSta",motors[mot].reg);
//getIntegerFromHardware(dev,&reg);
  return (reg & mask) ? TRUE : FALSE;
}
LIB_EXPORT int getSts(int mot,int *sts)
{ // is called from the background task @ 10 Hz
  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  *sts = motors[mot].status;
  return 0;
}
LIB_EXPORT int getGrpSts(int grpId,int *sts)
{
  int i, m;
  if (sts == NULL) return invalid_parameter;
  if (IsCoupledAxisMotor(grpId))
  { //TODO GROUP
    for (i=0; i<psAxGrps.numPsAxGroups; i++)
    {
      if (psAxGrps.grpIdXref[i] == grpId)
      { // got it
        *sts = psAxGrps.status[i];
        return 0;
      }
    }
  }
  else
  {
    for (i=0; i<ngrps; i++)
    {
      if (grps[i].id != grpId) continue;
      m = grps[i].grp.grphead - 1;
      *sts = motors[m].status;
      return 0;
    }
  }
  return invalid_index;
}
int makeMotStsBits(int mot)
{
  int stsBits=0;
  if (mot < 0 || mot > gNumMotors) return 0;
  if (motors[mot].idle) stsBits |= STS_IDLE;
  if (motors[mot].enabled) stsBits |= STS_ENABLED;
  if (motors[mot].moving) stsBits |= STS_MOVING;
  if (motors[mot].movpos) stsBits |= STS_MOVPOS;
  if (motors[mot].movneg) stsBits |= STS_MOVNEG;
  if (motors[mot].endpos) stsBits |= STS_ENDPOS;
  if (motors[mot].endneg) stsBits |= STS_ENDNEG;
  return stsBits;
}
LIB_EXPORT int getStsInfo(int mot,MotorStatus *motsts)
{
  if (motsts == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  strncpy(motsts->motName,motors[mot].name,64);
  motsts->stsBits = makeMotStsBits(mot);
  motsts->stsAxis = motors[mot].axisState;
  motsts->stsCoupled = motors[mot].coupledState;
  motsts->stsClbr = motors[mot].clbrSts;
  motsts->stateClbr = motors[mot].clbrState;
  motsts->idxCntlLoop = motors[mot].actCntrlLoopIndex;
  motsts->pos = (float)motors[mot].posRb;
  motsts->tgt = (float)motors[mot].tgtRb;
  motsts->lag = motors[mot].lag;
  motsts->vel = motors[mot].velRb;
  motsts->load = motors[mot].pcntMaxCur;
  motsts->ovrRide = motors[mot].overrideSts;
  return 0;
}

LIB_EXPORT int getLastHdwErr(int mot,NAME64 *errstr,int *errcode)
{
  char dev[128];
  int cc = 0;
  if (errstr == NULL && errcode == NULL) return invalid_parameter;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (errstr != NULL) 
  { 
    sprintf(dev,"%.32s.ErrorText",motors[mot].name);
    cc = getCharArrayFromHardware(dev,errstr->name,64);
    if (cc != 0) return cc;
  }
  if (errcode != NULL)
  { 
    sprintf(dev,"%.32s.ErrorId",motors[mot].name);
    cc = getIntegerFromHardware(dev,errcode);
    if (cc != 0) return cc;
  }
  return 0;
}
int startup(void)
{
  initialize();
  SetMotorList(mstNames,gNumMotors); // done
  SetStatusList(stsLst,sizStsLst);   // done 
  SetGetMotorStatusDispatch(getSts); // done
  SetGetGroupMotorStatusDispatch(getGrpSts);
  SetGetMotorVelocityDispatch(getVel); // done
  SetSetMotorVelocityDispatch(setVel); // done
  SetGetMotorAccelerationDispatch(getAcc);
  SetGetMotorPositionDispatch(getPos); // done
  SetGetGroupMotorPositionDispatch(getGrpPos);
  SetSetMotorPositionDispatch(setPos);
  SetGetMotorStepsPerMUDispatch(mstepsPerMM); // done
  SetGetMotorMicroStepsPerStepDispatch(mstepsPerStep); // done
  SetGetMotorStepsPerRadianDispatch(stepsPerRadian); // done
  SetGetMoveToStatusDispatch(movests); // done
  SetGetMoveToProgressDispatch(moveprg); // done
  SetGetGroupMoveStatusDispatch(movests);
  SetGetIncrementMoveStatusDispatch(movests);
  SetGetIncrementGroupMoveStatusDispatch(movests);
  SetSetMoveToStatusDispatch(moveto); // done
  SetSetFreeMoveToStatusDispatch(movetoNoEndSwitch); // done
  SetSetGroupMoveStatusDispatch(grpmoveto); // done ?
  SetSetIncrementGroupMoveStatusDispatch(grpmoveto);
  SetSetIncrementMoveStatusDispatch(moveto); // done (same as MoveTo)
  SetSetMotorTargetPositionDispatch(setTgtPos); // done
  SetGetMotorTargetPositionDispatch(getTgtPos); // done
  SetSetGroupMotorTargetPositionDispatch(setGrpTgtPos);
  SetGetGroupMotorTargetPositionDispatch(getGrpTgtPos);
  SetGetGroupMembersDispatch(getGrpMembers); // done ?
  SetSetGroupMembersDispatch(setGrpMembers); // done ?
  SetCalibrateMotorDispatch(calibr);  // done
  SetResetMotorDispatch(reset); // done
  SetGetMotorVelocityLowerLimitDispatch(getVelLwr);
  SetGetMotorVelocityUpperLimitDispatch(getVelUpr);
  SetGetMotorAccelerationUpperLimitDispatch(getAccUpr);
  SetGetMotorPositionMinimumDispatch(getPosMin); // done
  SetGetMotorPositionMaximumDispatch(getPosMax); // done
  SetGetMotorPositionOffsetDispatch(getPosOff); // done
  SetSetMotorPositionOffsetDispatch(setPosOff); // done
  SetGetMotorPositionScaleDispatch(getPosScale); // done
  SetSetMotorPositionScaleDispatch(setPosScale); // done
  SetCalibrateToOffsetDispatch(calibrToOffset);
  SetGetMotorUnitsDispatch(getUnits);
  SetGetScanMoveToStatusDispatch(scansts);
  SetSetScanMoveToStatusDispatch(scanto);
  SetGetScanGroupMoveToStatusDispatch(grpscansts);
  SetSetScanGroupMoveToStatusDispatch(grpscanto);
  SetGetScanMoveToProgressDispatch(scanprg);
  SetGetMotorOnlineDispatch(getonline);
  SetSetMotorOnlineDispatch(setonline);
  SetGetMotorStatusBlockDispatch(getStsInfo);
  SetSetGroupGearRatioDispatch(setGrpGearRatio); // done
  SetGetGroupGearRatioDispatch(getGrpGearRatio); // done
  SetGetLastHardwareErrorDispatch(getLastHdwErr);
  return 0;
}
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      LoadLibrary("cdi32.dll");
      SetInitMotorHardwareDispatch(startup);
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      break;
  }
  return (TRUE);
}
 